from collections.abc import Callable, Mapping, Sequence
from typing import Any, Literal

import dask.array as da
import numpy as np
import zarr
from pydantic import BaseModel, ConfigDict, model_validator

from ngio.common._zoom import (
    InterpolationOrder,
    _zoom_inputs_check,
    dask_zoom,
    numpy_zoom,
)
from ngio.utils import (
    NgioValueError,
)


def _on_disk_numpy_zoom(
    source: zarr.Array,
    target: zarr.Array,
    order: InterpolationOrder,
) -> None:
    source_array = source[...]
    if not isinstance(source_array, np.ndarray):
        raise NgioValueError("source zarr array could not be read as a numpy array")
    target[...] = numpy_zoom(source_array, target_shape=target.shape, order=order)


def _on_disk_dask_zoom(
    source: zarr.Array,
    target: zarr.Array,
    order: InterpolationOrder,
) -> None:
    source_array = da.from_zarr(source)
    target_array = dask_zoom(source_array, target_shape=target.shape, order=order)

    # This is a potential fix for Dask 2025.11
    # import dask.config
    # chunk_size_bytes = np.prod(target.chunks) * target_array.dtype.itemsize
    # current_chunk_size = dask.config.get("array.chunk-size")
    # Increase the chunk size to avoid dask potentially creating
    # corrupted chunks when writing chunks that are not multiple of the
    # target chunk size
    # dask.config.set({"array.chunk-size": f"{chunk_size_bytes}B"})
    target_array = target_array.rechunk(target.chunks)
    target_array = target_array.compute_chunk_sizes()
    target_array.to_zarr(target)
    # Restore previous chunk size
    # dask.config.set({"array.chunk-size": current_chunk_size})


def _on_disk_coarsen(
    source: zarr.Array,
    target: zarr.Array,
    order: InterpolationOrder = "linear",
    aggregation_function: Callable | None = None,
) -> None:
    """Apply a coarsening operation from a source zarr array to a target zarr array.

    Args:
        source (zarr.Array): The source array to coarsen.
        target (zarr.Array): The target array to save the coarsened result to.
        order (InterpolationOrder): The order of interpolation is not really implemented
            for coarsening, but it is kept for compatibility with the zoom function.
            order="linear" -> linear interpolation ~ np.mean
            order="nearest" -> nearest interpolation ~ np.max
        aggregation_function (np.ufunc): The aggregation function to use.
    """
    source_array = da.from_zarr(source)

    _scale, _target_shape = _zoom_inputs_check(
        source_array=source_array, scale=None, target_shape=target.shape
    )

    assert _target_shape == target.shape, (
        "Target shape must match the target array shape"
    )

    if aggregation_function is None:
        if order == "linear":
            aggregation_function = np.mean
        elif order == "nearest":
            aggregation_function = np.max
        elif order == "cubic":
            raise NgioValueError("Cubic interpolation is not supported for coarsening.")
        else:
            raise NgioValueError(
                f"Aggregation function must be provided for order {order}"
            )

    coarsening_setup = {}
    for i, s in enumerate(_scale):
        factor = 1 / s
        # This check is very strict, but it is necessary to avoid
        # a few pixels shift in the coarsening
        # We could add a tolerance
        if factor.is_integer():
            coarsening_setup[i] = int(factor)
        else:
            raise NgioValueError(
                f"Coarsening factor must be an integer, got {factor} on axis {i}"
            )

    out_target = da.coarsen(
        aggregation_function, source_array, coarsening_setup, trim_excess=True
    )
    out_target = out_target.rechunk(target.chunks)
    out_target.to_zarr(target)


def on_disk_zoom(
    source: zarr.Array,
    target: zarr.Array,
    order: InterpolationOrder = "linear",
    mode: Literal["dask", "numpy", "coarsen"] = "dask",
) -> None:
    """Apply a zoom operation from a source zarr array to a target zarr array.

    Args:
        source (zarr.Array): The source array to zoom.
        target (zarr.Array): The target array to save the zoomed result to.
        order (InterpolationOrder): The order of interpolation. Defaults to "linear".
        mode (Literal["dask", "numpy", "coarsen"]): The mode to use. Defaults to "dask".
    """
    if not isinstance(source, zarr.Array):
        raise NgioValueError("source must be a zarr array")

    if not isinstance(target, zarr.Array):
        raise NgioValueError("target must be a zarr array")

    if source.dtype != target.dtype:
        raise NgioValueError("source and target must have the same dtype")

    match mode:
        case "numpy":
            return _on_disk_numpy_zoom(source, target, order)
        case "dask":
            return _on_disk_dask_zoom(source, target, order)
        case "coarsen":
            return _on_disk_coarsen(
                source,
                target,
            )
        case _:
            raise NgioValueError("mode must be either 'dask', 'numpy' or 'coarsen'")


def _find_closest_arrays(
    processed: list[zarr.Array], to_be_processed: list[zarr.Array]
) -> tuple[np.intp, np.intp]:
    dist_matrix = np.zeros((len(processed), len(to_be_processed)))
    for i, arr_to_proc in enumerate(to_be_processed):
        for j, proc_arr in enumerate(processed):
            dist_matrix[j, i] = np.sqrt(
                np.sum(
                    [
                        (s1 - s2) ** 2
                        for s1, s2 in zip(
                            arr_to_proc.shape, proc_arr.shape, strict=False
                        )
                    ]
                )
            )

    indices = np.unravel_index(dist_matrix.argmin(), dist_matrix.shape)
    assert len(indices) == 2, "Indices must be of length 2"
    return indices


def consolidate_pyramid(
    source: zarr.Array,
    targets: list[zarr.Array],
    order: InterpolationOrder = "linear",
    mode: Literal["dask", "numpy", "coarsen"] = "dask",
) -> None:
    """Consolidate the Zarr array."""
    processed = [source]
    to_be_processed = targets

    while to_be_processed:
        source_id, target_id = _find_closest_arrays(processed, to_be_processed)

        source_image = processed[source_id]
        target_image = to_be_processed.pop(target_id)

        on_disk_zoom(
            source=source_image,
            target=target_image,
            mode=mode,
            order=order,
        )
        processed.append(target_image)


################################################
#
# Builders for image pyramids
#
################################################

ChunksLike = tuple[int, ...] | Literal["auto"]
ShardsLike = tuple[int, ...] | Literal["auto"]


def shapes_from_scaling_factors(
    base_shape: tuple[int, ...],
    scaling_factors: tuple[float, ...],
    num_levels: int,
) -> list[tuple[int, ...]]:
    """Compute the shapes of each level in the pyramid from scaling factors.

    Args:
        base_shape (tuple[int, ...]): The shape of the base level.
        scaling_factors (tuple[float, ...]): The scaling factors between levels.
        num_levels (int): The number of levels in the pyramid.

    Returns:
        list[tuple[int, ...]]: The shapes of each level in the pyramid.
    """
    shapes = []
    current_shape = base_shape
    for _ in range(num_levels):
        shapes.append(current_shape)
        current_shape = tuple(
            max(1, int(s / f))
            for s, f in zip(current_shape, scaling_factors, strict=True)
        )
    return shapes


def _check_order(shapes: Sequence[tuple[int, ...]]):
    """Check if the shapes are in decreasing order."""
    num_pixels = [np.prod(shape) for shape in shapes]
    for i in range(1, len(num_pixels)):
        if num_pixels[i] >= num_pixels[i - 1]:
            raise NgioValueError("Shapes are not in decreasing order.")


class PyramidLevel(BaseModel):
    path: str
    shape: tuple[int, ...]
    scale: tuple[float, ...]
    chunks: ChunksLike = "auto"
    shards: ShardsLike | None = None

    @model_validator(mode="after")
    def _model_validation(self) -> "PyramidLevel":
        # Same length as shape
        if len(self.scale) != len(self.shape):
            raise NgioValueError(
                "Scale must have the same length as shape "
                f"({len(self.shape)}), got {len(self.scale)}"
            )
        if any(isinstance(s, float) and s < 0 for s in self.scale):
            raise NgioValueError("Scale values must be positive.")

        if isinstance(self.chunks, tuple):
            if len(self.chunks) != len(self.shape):
                raise NgioValueError(
                    "Chunks must have the same length as shape "
                    f"({len(self.shape)}), got {len(self.chunks)}"
                )
            normalized_chunks = []
            for dim_size, chunk_size in zip(self.shape, self.chunks, strict=True):
                normalized_chunks.append(min(dim_size, chunk_size))
            self.chunks = tuple(normalized_chunks)

        if isinstance(self.shards, tuple):
            if len(self.shards) != len(self.shape):
                raise NgioValueError(
                    "Shards must have the same length as shape "
                    f"({len(self.shape)}), got {len(self.shards)}"
                )
        return self


class ImagePyramidBuilder(BaseModel):
    levels: list[PyramidLevel]
    axes: tuple[str, ...]
    data_type: str = "uint16"
    dimension_separator: Literal[".", "/"] = "/"
    compressors: Any = "auto"
    zarr_format: Literal[2, 3] = 2
    other_array_kwargs: Mapping[str, Any] = {}

    model_config = ConfigDict(arbitrary_types_allowed=True)

    @classmethod
    def from_scaling_factors(
        cls,
        levels_paths: tuple[str, ...],
        scaling_factors: tuple[float, ...],
        base_shape: tuple[int, ...],
        base_scale: tuple[float, ...],
        axes: tuple[str, ...],
        chunks: ChunksLike = "auto",
        shards: ShardsLike | None = None,
        data_type: str = "uint16",
        dimension_separator: Literal[".", "/"] = "/",
        compressors: Any = "auto",
        zarr_format: Literal[2, 3] = 2,
        other_array_kwargs: Mapping[str, Any] | None = None,
    ) -> "ImagePyramidBuilder":
        shapes = shapes_from_scaling_factors(
            base_shape=base_shape,
            scaling_factors=scaling_factors,
            num_levels=len(levels_paths),
        )
        return cls.from_shapes(
            shapes=shapes,
            base_scale=base_scale,
            axes=axes,
            levels_paths=levels_paths,
            chunks=chunks,
            shards=shards,
            data_type=data_type,
            dimension_separator=dimension_separator,
            compressors=compressors,
            zarr_format=zarr_format,
            other_array_kwargs=other_array_kwargs,
        )

    @classmethod
    def from_shapes(
        cls,
        shapes: Sequence[tuple[int, ...]],
        base_scale: tuple[float, ...],
        axes: tuple[str, ...],
        levels_paths: Sequence[str] | None = None,
        chunks: ChunksLike = "auto",
        shards: ShardsLike | None = None,
        data_type: str = "uint16",
        dimension_separator: Literal[".", "/"] = "/",
        compressors: Any = "auto",
        zarr_format: Literal[2, 3] = 2,
        other_array_kwargs: Mapping[str, Any] | None = None,
    ) -> "ImagePyramidBuilder":
        levels = []
        if levels_paths is None:
            levels_paths = tuple(str(i) for i in range(len(shapes)))
        _check_order(shapes)
        scale_ = base_scale
        for i, (path, shape) in enumerate(zip(levels_paths, shapes, strict=True)):
            levels.append(
                PyramidLevel(
                    path=path,
                    shape=shape,
                    scale=scale_,
                    chunks=chunks,
                    shards=shards,
                )
            )
            if i + 1 < len(shapes):
                # This only works for downsampling pyramids
                # The _check_order function ensures that
                # shapes are decreasing
                next_shape = shapes[i + 1]
                scaling_factor = tuple(
                    s1 / s2
                    for s1, s2 in zip(
                        shape,
                        next_shape,
                        strict=True,
                    )
                )
                scale_ = tuple(
                    s * f for s, f in zip(scale_, scaling_factor, strict=True)
                )
        other_array_kwargs = other_array_kwargs or {}
        return cls(
            levels=levels,
            axes=axes,
            data_type=data_type,
            dimension_separator=dimension_separator,
            compressors=compressors,
            zarr_format=zarr_format,
            other_array_kwargs=other_array_kwargs,
        )

    def to_zarr(self, group: zarr.Group) -> None:
        """Save the pyramid specification to a Zarr group.

        Args:
            group (zarr.Group): The Zarr group to save the pyramid specification to.
        """
        array_static_kwargs = {
            "dtype": self.data_type,
            "overwrite": True,
            "compressors": self.compressors,
            **self.other_array_kwargs,
        }

        if self.zarr_format == 2:
            array_static_kwargs["chunk_key_encoding"] = {
                "name": "v2",
                "separator": self.dimension_separator,
            }
        else:
            array_static_kwargs["chunk_key_encoding"] = {
                "name": "default",
                "separator": self.dimension_separator,
            }
            array_static_kwargs["dimension_names"] = self.axes
        for p_level in self.levels:
            group.create_array(
                name=p_level.path,
                shape=tuple(p_level.shape),
                chunks=p_level.chunks,
                shards=p_level.shards,
                **array_static_kwargs,
            )



from pathlib import Path

from utils import (
    check_ome_zarr,
    create_sample_ome_zarr,
    derive_image,
    get_http_mapper,
    get_s3_mapper,
    random_zarr_path,
)

from ngio import open_ome_zarr_container

HTTP_STORE_SUPPORTED_BACKENDS = ["anndata", "json", "csv", "parquet"]


def test_http_store(http_static_server: dict) -> None:
    # create boto3 client pointing at moto server
    url, root = http_static_server["url"], http_static_server["root"]

    zarr_path = random_zarr_path()
    local_store = root / zarr_path
    _ = create_sample_ome_zarr(
        store=local_store, supported_backends=HTTP_STORE_SUPPORTED_BACKENDS
    )
    http_mapper = get_http_mapper(url, zarr_path)
    ome_zarr = open_ome_zarr_container(store=http_mapper)
    check_ome_zarr(ome_zarr, supported_backends=HTTP_STORE_SUPPORTED_BACKENDS)


def test_http_store_derive_to_s3_store(
    http_static_server: dict, moto_s3_server: dict
) -> None:
    url, root = http_static_server["url"], http_static_server["root"]
    zarr_path = random_zarr_path()
    local_store = root / zarr_path
    _ = create_sample_ome_zarr(
        store=local_store, supported_backends=HTTP_STORE_SUPPORTED_BACKENDS
    )
    http_mapper = get_http_mapper(url, zarr_path)
    ome_zarr = open_ome_zarr_container(store=http_mapper)
    other_store = get_s3_mapper(
        base_url=moto_s3_server["endpoint_url"],
        bucket_name=moto_s3_server["bucket_name"],
        zarr_path=random_zarr_path(),
    )
    derived_ome_zarr = derive_image(ome_zarr, other_store=other_store)
    check_ome_zarr(derived_ome_zarr, supported_backends=HTTP_STORE_SUPPORTED_BACKENDS)


def test_http_store_derive_to_local_store(
    http_static_server: dict, tmp_path: Path
) -> None:
    url, root = http_static_server["url"], http_static_server["root"]
    zarr_path = random_zarr_path()
    local_store = root / zarr_path
    _ = create_sample_ome_zarr(
        store=local_store, supported_backends=HTTP_STORE_SUPPORTED_BACKENDS
    )
    http_mapper = get_http_mapper(url, zarr_path)
    ome_zarr = open_ome_zarr_container(store=http_mapper)

    other_store = tmp_path / "http_local_store_test" / random_zarr_path()
    derived_ome_zarr = derive_image(ome_zarr, other_store=other_store)
    check_ome_zarr(derived_ome_zarr, supported_backends=HTTP_STORE_SUPPORTED_BACKENDS)


def test_http_store_derive_to_memory_store(http_static_server: dict) -> None:
    url, root = http_static_server["url"], http_static_server["root"]
    zarr_path = random_zarr_path()
    local_store = root / zarr_path
    _ = create_sample_ome_zarr(
        store=local_store, supported_backends=HTTP_STORE_SUPPORTED_BACKENDS
    )
    http_mapper = get_http_mapper(url, zarr_path)
    ome_zarr = open_ome_zarr_container(store=http_mapper)

    other_store = {}
    derived_ome_zarr = derive_image(ome_zarr, other_store=other_store)
    check_ome_zarr(derived_ome_zarr, supported_backends=["anndata", "json"])



"""Common classes and functions that are used across the package."""

from ngio.common._dimensions import Dimensions
from ngio.common._masking_roi import compute_masking_roi
from ngio.common._pyramid import (
    ChunksLike,
    ImagePyramidBuilder,
    ShardsLike,
    consolidate_pyramid,
    on_disk_zoom,
)
from ngio.common._roi import Roi, RoiSlice
from ngio.common._zoom import InterpolationOrder, dask_zoom, numpy_zoom

__all__ = [
    "ChunksLike",
    "Dimensions",
    "ImagePyramidBuilder",
    "InterpolationOrder",
    "Roi",
    "RoiSlice",
    "ShardsLike",
    "compute_masking_roi",
    "consolidate_pyramid",
    "dask_zoom",
    "numpy_zoom",
    "on_disk_zoom",
]



"""OME-Zarr HCS objects models."""

from ngio.hcs._plate import (
    OmeZarrPlate,
    OmeZarrWell,
    create_empty_plate,
    create_empty_well,
    open_ome_zarr_plate,
    open_ome_zarr_well,
)

__all__ = [
    "OmeZarrPlate",
    "OmeZarrWell",
    "create_empty_plate",
    "create_empty_well",
    "open_ome_zarr_plate",
    "open_ome_zarr_well",
]



from typing import Annotated

from ome_zarr_models.v05.well import WellAttrs as WellAttrs05
from ome_zarr_models.v05.well_types import WellImage as WellImage05
from ome_zarr_models.v05.well_types import WellMeta as WellMeta05
from pydantic import SkipValidation


class CustomWellImage(WellImage05):
    path: Annotated[str, SkipValidation]


class CustomWellMeta(WellMeta05):
    images: list[CustomWellImage]  # type: ignore[valid-type]


class CustomWellAttrs(WellAttrs05):
    well: CustomWellMeta  # type: ignore[valid-type]



from pathlib import Path

import pytest

from ngio.tables._tables_container import open_table, write_table
from ngio.tables.v1._roi_table import MaskingRoiTableV1, Roi
from ngio.utils import NgioValueError


def test_masking_roi_table_v1(tmp_path: Path):
    rois = {
        1: Roi.from_values(
            name="1",
            slices={"x": slice(0, 10), "y": slice(0, 10), "z": slice(0, 5)},
        )
    }

    table = MaskingRoiTableV1(rois=rois.values(), reference_label="label")
    assert isinstance(table.__repr__(), str)
    assert table.reference_label == "label"
    assert table.meta.region is not None
    assert table.meta.region.path == "../labels/label"

    table.add(
        roi=Roi.from_values(
            name="2",
            slices={"x": slice(0, 10), "y": slice(0, 10), "z": slice(0, 5)},
        )
    )

    with pytest.raises(NgioValueError):
        table.add(
            roi=Roi.from_values(
                name="2",
                slices={"x": slice(0, 10), "y": slice(0, 10), "z": slice(0, 5)},
            )
        )

    write_table(store=tmp_path / "roi_table.zarr", table=table, backend="anndata")

    loaded_table = open_table(store=tmp_path / "roi_table.zarr")
    assert isinstance(loaded_table, MaskingRoiTableV1)

    assert loaded_table.meta.backend == "anndata"
    meta_dict = loaded_table._meta.model_dump()
    assert meta_dict.get("table_version") == loaded_table.version()
    assert meta_dict.get("type") == loaded_table.table_type()



"""Base class for handling OME-NGFF metadata in Zarr groups."""

from collections.abc import Callable
from typing import TypeVar

from ngio.ome_zarr_meta.ngio_specs import (
    AxesSetup,
    NgioImageMeta,
    NgioLabelMeta,
    NgioLabelsGroupMeta,
    NgioPlateMeta,
    NgioWellMeta,
)
from ngio.ome_zarr_meta.ngio_specs._ngio_image import NgffVersions
from ngio.ome_zarr_meta.v04 import (
    ngio_to_v04_image_meta,
    ngio_to_v04_label_meta,
    ngio_to_v04_labels_group_meta,
    ngio_to_v04_plate_meta,
    ngio_to_v04_well_meta,
    v04_to_ngio_image_meta,
    v04_to_ngio_label_meta,
    v04_to_ngio_labels_group_meta,
    v04_to_ngio_plate_meta,
    v04_to_ngio_well_meta,
)
from ngio.ome_zarr_meta.v05 import (
    ngio_to_v05_image_meta,
    ngio_to_v05_label_meta,
    ngio_to_v05_labels_group_meta,
    ngio_to_v05_plate_meta,
    ngio_to_v05_well_meta,
    v05_to_ngio_image_meta,
    v05_to_ngio_label_meta,
    v05_to_ngio_labels_group_meta,
    v05_to_ngio_plate_meta,
    v05_to_ngio_well_meta,
)
from ngio.utils import (
    NgioValidationError,
    NgioValueError,
    ZarrGroupHandler,
)

# This could be replaced with a more dynamic registry if needed in the future
_image_encoder_registry = {"0.4": ngio_to_v04_image_meta, "0.5": ngio_to_v05_image_meta}
_image_decoder_registry = {"0.4": v04_to_ngio_image_meta, "0.5": v05_to_ngio_image_meta}
_label_encoder_registry = {"0.4": ngio_to_v04_label_meta, "0.5": ngio_to_v05_label_meta}
_label_decoder_registry = {"0.4": v04_to_ngio_label_meta, "0.5": v05_to_ngio_label_meta}
_plate_encoder_registry = {"0.4": ngio_to_v04_plate_meta, "0.5": ngio_to_v05_plate_meta}
_plate_decoder_registry = {"0.4": v04_to_ngio_plate_meta, "0.5": v05_to_ngio_plate_meta}
_well_encoder_registry = {"0.4": ngio_to_v04_well_meta, "0.5": ngio_to_v05_well_meta}
_well_decoder_registry = {"0.4": v04_to_ngio_well_meta, "0.5": v05_to_ngio_well_meta}
_labels_group_encoder_registry = {
    "0.4": ngio_to_v04_labels_group_meta,
    "0.5": ngio_to_v05_labels_group_meta,
}
_labels_group_decoder_registry = {
    "0.4": v04_to_ngio_labels_group_meta,
    "0.5": v05_to_ngio_labels_group_meta,
}

_meta_type = TypeVar(
    "_meta_type",
    NgioImageMeta,
    NgioLabelMeta,
    NgioLabelsGroupMeta,
    NgioPlateMeta,
    NgioWellMeta,
)


def _find_encoder_registry(
    ngio_meta: _meta_type,
) -> dict[str, Callable]:
    if isinstance(ngio_meta, NgioImageMeta):
        return _image_encoder_registry
    elif isinstance(ngio_meta, NgioLabelMeta):
        return _label_encoder_registry
    elif isinstance(ngio_meta, NgioPlateMeta):
        return _plate_encoder_registry
    elif isinstance(ngio_meta, NgioWellMeta):
        return _well_encoder_registry
    elif isinstance(ngio_meta, NgioLabelsGroupMeta):
        return _labels_group_encoder_registry
    else:
        raise NgioValueError(f"Unsupported NGIO metadata type: {type(ngio_meta)}")


def update_ngio_meta(
    group_handler: ZarrGroupHandler,
    ngio_meta: _meta_type,
) -> None:
    """Update the metadata in the Zarr group.

    Args:
        group_handler (ZarrGroupHandler): The Zarr group handler.
        ngio_meta (_meta_type): The new NGIO metadata.

    """
    registry = _find_encoder_registry(ngio_meta)
    exporter = registry.get(ngio_meta.version)
    if exporter is None:
        raise NgioValueError(f"Unsupported NGFF version: {ngio_meta.version}")

    zarr_meta = exporter(ngio_meta)
    group_handler.write_attrs(zarr_meta)


def _find_decoder_registry(
    meta_type: type[_meta_type],
) -> dict[str, Callable]:
    if meta_type is NgioImageMeta:
        return _image_decoder_registry
    elif meta_type is NgioLabelMeta:
        return _label_decoder_registry
    elif meta_type is NgioPlateMeta:
        return _plate_decoder_registry
    elif meta_type is NgioWellMeta:
        return _well_decoder_registry
    elif meta_type is NgioLabelsGroupMeta:
        return _labels_group_decoder_registry
    else:
        raise NgioValueError(f"Unsupported NGIO metadata type: {meta_type}")


def get_ngio_meta(
    group_handler: ZarrGroupHandler,
    meta_type: type[_meta_type],
    version: str | None = None,
    **kwargs,
) -> _meta_type:
    """Retrieve the NGIO metadata from the Zarr group.

    Args:
        group_handler (ZarrGroupHandler): The Zarr group handler.
        meta_type (type[_meta_type]): The type of NGIO metadata to retrieve.
        version (str | None): Optional NGFF version to use for decoding.
        **kwargs: Additional arguments to pass to the decoder.

    Returns:
        _meta_type: The NGIO metadata.
    """
    registry = _find_decoder_registry(meta_type)
    if version is not None:
        decoder = registry.get(version)
        if decoder is None:
            raise NgioValueError(f"Unsupported NGFF version: {version}")
        versions_to_try = {version: decoder}
    else:
        versions_to_try = registry

    attrs = group_handler.load_attrs()
    all_errors = []
    for version, decoder in versions_to_try.items():
        try:
            ngio_meta = decoder(attrs, **kwargs)
            return ngio_meta
        except Exception as e:
            all_errors.append(f"Version {version}: {e}")
    error_message = (
        f"Failed to decode NGIO {meta_type.__name__} metadata:\n"
        + "\n".join(all_errors)
    )
    raise NgioValidationError(error_message)


##################################################
#
# Concrete implementations for NGIO metadata types
#
##################################################


def get_ngio_image_meta(
    group_handler: ZarrGroupHandler,
    version: str | None = None,
    axes_setup: AxesSetup | None = None,
    allow_non_canonical_axes: bool = False,
    strict_canonical_order: bool = True,
) -> NgioImageMeta:
    """Retrieve the NGIO image metadata from the Zarr group.

    Args:
        group_handler (ZarrGroupHandler): The Zarr group handler.
        version (str | None): Optional NGFF version to use for decoding.
        axes_setup (AxesSetup | None): Optional axes setup for validation.
        allow_non_canonical_axes (bool): Whether to allow non-canonical axes.
        strict_canonical_order (bool): Whether to enforce strict canonical order.

    Returns:
        NgioImageMeta: The NGIO image metadata.
    """
    return get_ngio_meta(
        group_handler=group_handler,
        meta_type=NgioImageMeta,
        version=version,
        axes_setup=axes_setup,
        allow_non_canonical_axes=allow_non_canonical_axes,
        strict_canonical_order=strict_canonical_order,
    )


def update_ngio_image_meta(
    group_handler: ZarrGroupHandler,
    ngio_meta: NgioImageMeta,
) -> None:
    """Update the NGIO image metadata in the Zarr group.

    Args:
        group_handler (ZarrGroupHandler): The Zarr group handler.
        ngio_meta (NgioImageMeta): The new NGIO image metadata.

    """
    update_ngio_meta(
        group_handler=group_handler,
        ngio_meta=ngio_meta,
    )


class ImageMetaHandler:
    def __init__(
        self,
        group_handler: ZarrGroupHandler,
        version: str | None = None,
        axes_setup: AxesSetup | None = None,
        allow_non_canonical_axes: bool = False,
        strict_canonical_order: bool = True,
    ):
        self._group_handler = group_handler
        self._version = version
        self._axes_setup = axes_setup
        self._allow_non_canonical_axes = allow_non_canonical_axes
        self._strict_canonical_order = strict_canonical_order

        # Validate metadata
        meta = self.get_meta()
        # Store the resolved version
        self._version = meta.version

    def get_meta(self) -> NgioImageMeta:
        """Retrieve the NGIO image metadata."""
        return get_ngio_image_meta(
            group_handler=self._group_handler,
            version=self._version,
            axes_setup=self._axes_setup,
            allow_non_canonical_axes=self._allow_non_canonical_axes,
            strict_canonical_order=self._strict_canonical_order,
        )

    def update_meta(self, ngio_meta: NgioImageMeta) -> None:
        """Update the NGIO image metadata."""
        update_ngio_meta(
            group_handler=self._group_handler,
            ngio_meta=ngio_meta,
        )


def get_ngio_label_meta(
    group_handler: ZarrGroupHandler,
    version: str | None = None,
    axes_setup: AxesSetup | None = None,
    allow_non_canonical_axes: bool = False,
    strict_canonical_order: bool = True,
) -> NgioLabelMeta:
    """Retrieve the NGIO label metadata from the Zarr group.

    Args:
        group_handler (ZarrGroupHandler): The Zarr group handler.
        version (str | None): Optional NGFF version to use for decoding.
        axes_setup (AxesSetup | None): Optional axes setup for validation.
        allow_non_canonical_axes (bool): Whether to allow non-canonical axes.
        strict_canonical_order (bool): Whether to enforce strict canonical order.

    Returns:
        NgioLabelMeta: The NGIO label metadata.
    """
    return get_ngio_meta(
        group_handler=group_handler,
        meta_type=NgioLabelMeta,
        version=version,
        axes_setup=axes_setup,
        allow_non_canonical_axes=allow_non_canonical_axes,
        strict_canonical_order=strict_canonical_order,
    )


def update_ngio_label_meta(
    group_handler: ZarrGroupHandler,
    ngio_meta: NgioLabelMeta,
) -> None:
    """Update the NGIO label metadata in the Zarr group.

    Args:
        group_handler (ZarrGroupHandler): The Zarr group handler.
        ngio_meta (NgioLabelMeta): The new NGIO label metadata.

    """
    update_ngio_meta(
        group_handler=group_handler,
        ngio_meta=ngio_meta,
    )


class LabelMetaHandler:
    def __init__(
        self,
        group_handler: ZarrGroupHandler,
        version: str | None = None,
        axes_setup: AxesSetup | None = None,
        allow_non_canonical_axes: bool = False,
        strict_canonical_order: bool = True,
    ):
        self._group_handler = group_handler
        self._version = version
        self._axes_setup = axes_setup
        self._allow_non_canonical_axes = allow_non_canonical_axes
        self._strict_canonical_order = strict_canonical_order

        # Validate metadata
        meta = self.get_meta()
        # Store the resolved version
        self._version = meta.version

    def get_meta(self) -> NgioLabelMeta:
        """Retrieve the NGIO label metadata."""
        return get_ngio_label_meta(
            group_handler=self._group_handler,
            version=self._version,
            axes_setup=self._axes_setup,
            allow_non_canonical_axes=self._allow_non_canonical_axes,
            strict_canonical_order=self._strict_canonical_order,
        )

    def update_meta(self, ngio_meta: NgioLabelMeta) -> None:
        """Update the NGIO label metadata."""
        update_ngio_meta(
            group_handler=self._group_handler,
            ngio_meta=ngio_meta,
        )


def get_ngio_plate_meta(
    group_handler: ZarrGroupHandler,
    version: str | None = None,
) -> NgioPlateMeta:
    """Retrieve the NGIO plate metadata from the Zarr group.

    Args:
        group_handler (ZarrGroupHandler): The Zarr group handler.
        version (str | None): Optional NGFF version to use for decoding.

    Returns:
        NgioPlateMeta: The NGIO plate metadata.
    """
    return get_ngio_meta(
        group_handler=group_handler,
        meta_type=NgioPlateMeta,
        version=version,
    )


def update_ngio_plate_meta(
    group_handler: ZarrGroupHandler,
    ngio_meta: NgioPlateMeta,
) -> None:
    """Update the NGIO plate metadata in the Zarr group.

    Args:
        group_handler (ZarrGroupHandler): The Zarr group handler.
        ngio_meta (NgioPlateMeta): The new NGIO plate metadata.

    """
    update_ngio_meta(
        group_handler=group_handler,
        ngio_meta=ngio_meta,
    )


class PlateMetaHandler:
    def __init__(
        self,
        group_handler: ZarrGroupHandler,
        version: str | None = None,
    ):
        self._group_handler = group_handler
        self._version = version

        # Validate metadata
        _ = self.get_meta()
        # Store the resolved version
        # self._version = meta.version

    def get_meta(self) -> NgioPlateMeta:
        """Retrieve the NGIO plate metadata."""
        return get_ngio_plate_meta(
            group_handler=self._group_handler,
            version=self._version,
        )

    def update_meta(self, ngio_meta: NgioPlateMeta) -> None:
        """Update the NGIO plate metadata."""
        update_ngio_meta(
            group_handler=self._group_handler,
            ngio_meta=ngio_meta,
        )


def get_ngio_well_meta(
    group_handler: ZarrGroupHandler,
    version: str | None = None,
) -> NgioWellMeta:
    """Retrieve the NGIO well metadata from the Zarr group.

    Args:
        group_handler (ZarrGroupHandler): The Zarr group handler.
        version (str | None): Optional NGFF version to use for decoding.

    Returns:
        NgioWellMeta: The NGIO well metadata.
    """
    return get_ngio_meta(
        group_handler=group_handler,
        meta_type=NgioWellMeta,
        version=version,
    )


def update_ngio_well_meta(
    group_handler: ZarrGroupHandler,
    ngio_meta: NgioWellMeta,
) -> None:
    """Update the NGIO well metadata in the Zarr group.

    Args:
        group_handler (ZarrGroupHandler): The Zarr group handler.
        ngio_meta (NgioWellMeta): The new NGIO well metadata.

    """
    update_ngio_meta(
        group_handler=group_handler,
        ngio_meta=ngio_meta,
    )


class WellMetaHandler:
    def __init__(
        self,
        group_handler: ZarrGroupHandler,
        version: str | None = None,
    ):
        self._group_handler = group_handler
        self._version = version

        # Validate metadata
        _ = self.get_meta()
        # Store the resolved version
        # self._version = meta.version

    def get_meta(self) -> NgioWellMeta:
        """Retrieve the NGIO well metadata."""
        return get_ngio_well_meta(
            group_handler=self._group_handler,
            version=self._version,
        )

    def update_meta(self, ngio_meta: NgioWellMeta) -> None:
        """Update the NGIO well metadata."""
        update_ngio_meta(
            group_handler=self._group_handler,
            ngio_meta=ngio_meta,
        )


def get_ngio_labels_group_meta(
    group_handler: ZarrGroupHandler,
    version: str | None = None,
) -> NgioLabelsGroupMeta:
    """Retrieve the NGIO labels group metadata from the Zarr group.

    Args:
        group_handler (ZarrGroupHandler): The Zarr group handler.
        version (str | None): Optional NGFF version to use for decoding.

    Returns:
        NgioLabelsGroupMeta: The NGIO labels group metadata.
    """
    return get_ngio_meta(
        group_handler=group_handler,
        meta_type=NgioLabelsGroupMeta,
        version=version,
    )


def update_ngio_labels_group_meta(
    group_handler: ZarrGroupHandler,
    ngio_meta: NgioLabelsGroupMeta,
) -> None:
    """Update the NGIO labels group metadata in the Zarr group.

    Args:
        group_handler (ZarrGroupHandler): The Zarr group handler.
        ngio_meta (NgioLabelsGroupMeta): The new NGIO labels group metadata.

    """
    update_ngio_meta(
        group_handler=group_handler,
        ngio_meta=ngio_meta,
    )


class LabelsGroupMetaHandler:
    def __init__(
        self,
        group_handler: ZarrGroupHandler,
        version: NgffVersions | None = None,
    ):
        self._group_handler = group_handler
        self._version = version

        meta = self.get_meta()
        self._version = meta.version

    def get_meta(self) -> NgioLabelsGroupMeta:
        """Retrieve the NGIO labels group metadata."""
        return get_ngio_labels_group_meta(
            group_handler=self._group_handler,
            version=self._version,
        )

    def update_meta(self, ngio_meta: NgioLabelsGroupMeta) -> None:
        """Update the NGIO labels group metadata."""
        update_ngio_labels_group_meta(
            group_handler=self._group_handler,
            ngio_meta=ngio_meta,
        )



"""Utilities for OME-Zarr v05 specs.

This module provides a set of classes to internally handle the metadata
of the OME-Zarr v05 specification.

For Images and Labels implements the following functionalities:
- A function to find if a dict view of the metadata is a valid OME-Zarr v05 metadata.
- A function to convert a v05 image metadata to a ngio image metadata.
- A function to convert a ngio image metadata to a v05 image metadata.
"""

from ome_zarr_models.common.omero import Channel as ChannelV05
from ome_zarr_models.common.omero import Omero as OmeroV05
from ome_zarr_models.common.omero import Window as WindowV05
from ome_zarr_models.v05.axes import Axis as AxisV05
from ome_zarr_models.v05.coordinate_transformations import VectorScale as VectorScaleV05
from ome_zarr_models.v05.coordinate_transformations import (
    VectorTranslation as VectorTranslationV05,
)
from ome_zarr_models.v05.hcs import HCSAttrs as HCSAttrsV05
from ome_zarr_models.v05.image import ImageAttrs as ImageAttrsV05
from ome_zarr_models.v05.image_label import ImageLabelAttrs as ImageLabelAttrsV05
from ome_zarr_models.v05.labels import Labels as Labels
from ome_zarr_models.v05.labels import LabelsAttrs as LabelsAttrsV05
from ome_zarr_models.v05.multiscales import Dataset as DatasetV05
from ome_zarr_models.v05.multiscales import Multiscale as MultiscaleV05
from ome_zarr_models.v05.multiscales import ValidTransform as ValidTransformV05
from pydantic import BaseModel

from ngio.ome_zarr_meta.ngio_specs import (
    AxesHandler,
    AxesSetup,
    Axis,
    AxisType,
    Channel,
    ChannelsMeta,
    ChannelVisualisation,
    Dataset,
    ImageLabelSource,
    NgioImageMeta,
    NgioLabelMeta,
    NgioLabelsGroupMeta,
    NgioPlateMeta,
    NgioWellMeta,
    default_channel_name,
)
from ngio.ome_zarr_meta.v05._custom_models import CustomWellAttrs as WellAttrsV05


class ImageV05AttrsWithOmero(ImageAttrsV05):
    omero: OmeroV05 | None = None


class ImageV05WithOmero(BaseModel):
    ome: ImageV05AttrsWithOmero


class ImageLabelV05(BaseModel):
    ome: ImageLabelAttrsV05


def _v05_omero_to_channels(v05_omero: OmeroV05 | None) -> ChannelsMeta | None:
    if v05_omero is None:
        return None

    ngio_channels = []
    for idx, v05_channel in enumerate(v05_omero.channels):
        channel_extra = v05_channel.model_extra

        if channel_extra is None:
            channel_extra = {}

        if "label" in channel_extra:
            label = channel_extra.pop("label")
        else:
            label = default_channel_name(idx)

        if "wavelength_id" in channel_extra:
            wavelength_id = channel_extra.pop("wavelength_id")
        else:
            wavelength_id = label

        if "active" in channel_extra:
            active = channel_extra.pop("active")
        else:
            active = True

        channel_visualisation = ChannelVisualisation(
            color=v05_channel.color,
            start=v05_channel.window.start,
            end=v05_channel.window.end,
            min=v05_channel.window.min,
            max=v05_channel.window.max,
            active=active,
            **channel_extra,
        )

        ngio_channels.append(
            Channel(
                label=label,
                wavelength_id=wavelength_id,
                channel_visualisation=channel_visualisation,
            )
        )

    v05_omero_extra = v05_omero.model_extra if v05_omero.model_extra is not None else {}
    return ChannelsMeta(channels=ngio_channels, **v05_omero_extra)


def _compute_scale_translation(
    v05_transforms: ValidTransformV05,
    scale: list[float],
    translation: list[float],
) -> tuple[list[float], list[float]]:
    for v05_transform in v05_transforms:
        if isinstance(v05_transform, VectorScaleV05):
            scale = [t1 * t2 for t1, t2 in zip(scale, v05_transform.scale, strict=True)]

        elif isinstance(v05_transform, VectorTranslationV05):
            translation = [
                t1 + t2
                for t1, t2 in zip(translation, v05_transform.translation, strict=True)
            ]
        else:
            raise NotImplementedError(
                f"Coordinate transformation {v05_transform} is not supported."
            )
    return scale, translation


def _v05_to_ngio_datasets(
    v05_multiscale: MultiscaleV05,
    axes_setup: AxesSetup,
    allow_non_canonical_axes: bool = False,
    strict_canonical_order: bool = True,
) -> list[Dataset]:
    """Convert a v05 multiscale to a list of ngio datasets."""
    datasets = []

    global_scale = [1.0] * len(v05_multiscale.axes)
    global_translation = [0.0] * len(v05_multiscale.axes)

    if v05_multiscale.coordinateTransformations is not None:
        global_scale, global_translation = _compute_scale_translation(
            v05_multiscale.coordinateTransformations, global_scale, global_translation
        )

    # Prepare axes handler
    axes = []
    for v05_axis in v05_multiscale.axes:
        unit = v05_axis.unit
        if unit is not None and not isinstance(unit, str):
            unit = str(unit)
        axes.append(
            Axis(
                name=str(v05_axis.name),
                axis_type=AxisType(v05_axis.type),
                # (for some reason the type is a generic JsonValue,
                # but it should be a string or None)
                unit=v05_axis.unit,  # type: ignore
            )
        )
    axes_handler = AxesHandler(
        axes=axes,
        axes_setup=axes_setup,
        allow_non_canonical_axes=allow_non_canonical_axes,
        strict_canonical_order=strict_canonical_order,
    )

    for v05_dataset in v05_multiscale.datasets:
        _scale, _translation = _compute_scale_translation(
            v05_dataset.coordinateTransformations, global_scale, global_translation
        )
        datasets.append(
            Dataset(
                path=v05_dataset.path,
                axes_handler=axes_handler,
                scale=_scale,
                translation=_translation,
            )
        )
    return datasets


def v05_to_ngio_image_meta(
    metadata: dict,
    axes_setup: AxesSetup | None = None,
    allow_non_canonical_axes: bool = False,
    strict_canonical_order: bool = True,
) -> NgioImageMeta:
    """Convert a v05 image metadata to a ngio image metadata.

    Args:
        metadata (dict): The v05 image metadata.
        axes_setup (AxesSetup, optional): The axes setup. This is
            required to convert image with non-canonical axes names.
        allow_non_canonical_axes (bool, optional): Allow non-canonical axes.
        strict_canonical_order (bool, optional): Strict canonical order.

    Returns:
        NgioImageMeta: The ngio image metadata.
    """
    v05_image = ImageV05WithOmero(**metadata)
    v05_image = v05_image.ome
    if len(v05_image.multiscales) > 1:
        raise NotImplementedError(
            "Multiple multiscales in a single image are not supported in ngio."
        )

    v05_multiscale = v05_image.multiscales[0]

    channels_meta = _v05_omero_to_channels(v05_image.omero)
    axes_setup = axes_setup if axes_setup is not None else AxesSetup()
    datasets = _v05_to_ngio_datasets(
        v05_multiscale,
        axes_setup=axes_setup,
        allow_non_canonical_axes=allow_non_canonical_axes,
        strict_canonical_order=strict_canonical_order,
    )

    name = v05_multiscale.name
    if name is not None and not isinstance(name, str):
        name = str(name)
    return NgioImageMeta(
        version="0.5",
        name=name,
        datasets=datasets,
        channels=channels_meta,
    )


def v05_to_ngio_label_meta(
    metadata: dict,
    axes_setup: AxesSetup | None = None,
    allow_non_canonical_axes: bool = False,
    strict_canonical_order: bool = True,
) -> NgioLabelMeta:
    """Convert a v05 image metadata to a ngio image metadata.

    Args:
        metadata (dict): The v05 image metadata.
        axes_setup (AxesSetup, optional): The axes setup. This is
            required to convert image with non-canonical axes names.
        allow_non_canonical_axes (bool, optional): Allow non-canonical axes.
        strict_canonical_order (bool, optional): Strict canonical order.

    Returns:
        NgioLabelMeta: The ngio label metadata.
    """
    v05_label = ImageLabelV05(**metadata)
    v05_label = v05_label.ome

    if len(v05_label.multiscales) > 1:
        raise NotImplementedError(
            "Multiple multiscales in a single image are not supported in ngio."
        )

    v05_multiscale = v05_label.multiscales[0]

    axes_setup = axes_setup if axes_setup is not None else AxesSetup()
    datasets = _v05_to_ngio_datasets(
        v05_multiscale,
        axes_setup=axes_setup,
        allow_non_canonical_axes=allow_non_canonical_axes,
        strict_canonical_order=strict_canonical_order,
    )

    if v05_label.image_label is not None:
        source = v05_label.image_label.source
        if source is None:
            image_label_source = None
        else:
            source = v05_label.image_label.source
            if source is None:
                image_label_source = None
            else:
                image_label_source = source.image
            image_label_source = ImageLabelSource(
                version="0.5",
                source={"image": image_label_source},
            )
    else:
        image_label_source = None
    name = v05_multiscale.name
    if name is not None and not isinstance(name, str):
        name = str(name)

    return NgioLabelMeta(
        version="0.5",
        name=name,
        datasets=datasets,
        image_label=image_label_source,
    )


def _ngio_to_v05_multiscale(name: str | None, datasets: list[Dataset]) -> MultiscaleV05:
    """Convert a ngio multiscale to a v05 multiscale.

    Args:
        name (str | None): The name of the multiscale.
        datasets (list[Dataset]): The ngio datasets.

    Returns:
        MultiscaleV05: The v05 multiscale.
    """
    ax_mapper = datasets[0].axes_handler
    v05_axes = []
    for axis in ax_mapper.axes:
        v05_axes.append(
            AxisV05(
                name=axis.name,
                type=axis.axis_type.value if axis.axis_type is not None else None,
                unit=axis.unit if axis.unit is not None else None,
            )
        )

    v05_datasets = []
    for dataset in datasets:
        transform = [VectorScaleV05(type="scale", scale=list(dataset._scale))]
        if sum(dataset._translation) > 0:
            transform = (
                VectorScaleV05(type="scale", scale=list(dataset._scale)),
                VectorTranslationV05(
                    type="translation", translation=list(dataset._translation)
                ),
            )
        else:
            transform = (VectorScaleV05(type="scale", scale=list(dataset._scale)),)

        v05_datasets.append(
            DatasetV05(path=dataset.path, coordinateTransformations=transform)
        )
    return MultiscaleV05(axes=v05_axes, datasets=tuple(v05_datasets), name=name)


def _ngio_to_v05_omero(channels: ChannelsMeta | None) -> OmeroV05 | None:
    """Convert a ngio channels to a v05 omero."""
    if channels is None:
        return None

    v05_channels = []
    for channel in channels.channels:
        _model_extra = {
            "label": channel.label,
            "wavelength_id": channel.wavelength_id,
            "active": channel.channel_visualisation.active,
        }
        if channel.channel_visualisation.model_extra is not None:
            _model_extra.update(channel.channel_visualisation.model_extra)

        v05_channels.append(
            ChannelV05(
                color=channel.channel_visualisation.valid_color,
                window=WindowV05(
                    start=channel.channel_visualisation.start,
                    end=channel.channel_visualisation.end,
                    min=channel.channel_visualisation.min,
                    max=channel.channel_visualisation.max,
                ),
                **_model_extra,
            )
        )

    _model_extra = channels.model_extra if channels.model_extra is not None else {}
    return OmeroV05(channels=v05_channels, **_model_extra)


def ngio_to_v05_image_meta(metadata: NgioImageMeta) -> dict:
    """Convert a ngio image metadata to a v05 image metadata.

    Args:
        metadata (NgioImageMeta): The ngio image metadata.

    Returns:
        dict: The v05 image metadata.
    """
    v05_muliscale = _ngio_to_v05_multiscale(
        name=metadata.name, datasets=metadata.datasets
    )
    v05_omero = _ngio_to_v05_omero(metadata._channels_meta)

    v05_image_attrs = ImageV05AttrsWithOmero(
        multiscales=[v05_muliscale], omero=v05_omero, version="0.5"
    )
    v05_image = ImageV05WithOmero(
        ome=v05_image_attrs,
    )
    return v05_image.model_dump(exclude_none=True, by_alias=True)


def ngio_to_v05_label_meta(metadata: NgioLabelMeta) -> dict:
    """Convert a ngio image metadata to a v05 image metadata.

    Args:
        metadata (NgioImageMeta): The ngio image metadata.

    Returns:
        dict: The v05 image metadata.
    """
    v05_muliscale = _ngio_to_v05_multiscale(
        name=metadata.name, datasets=metadata.datasets
    )
    labels_meta = {
        "multiscales": [v05_muliscale],
        "image-label": metadata.image_label.model_dump(),
    }
    v05_label = ImageLabelAttrsV05(**labels_meta, version="0.5")
    v05_label = ImageLabelV05(
        ome=v05_label,
    )
    return v05_label.model_dump(exclude_none=True, by_alias=True)


class WellV05(BaseModel):
    ome: WellAttrsV05


class HCSV05(BaseModel):
    ome: HCSAttrsV05


def v05_to_ngio_well_meta(
    metadata: dict,
) -> NgioWellMeta:
    """Convert a v05 well metadata to a ngio well metadata.

    Args:
        metadata (dict): The v05 well metadata.

    Returns:
        NgioWellMeta: The ngio well metadata.
    """
    v05_well = WellV05(**metadata).ome.well.model_dump()
    images = v05_well.get("images", [])
    return NgioWellMeta(images=images, version="0.5")


def v05_to_ngio_plate_meta(
    metadata: dict,
) -> NgioPlateMeta:
    """Convert a v05 plate metadata to a ngio plate metadata.

    Args:
        metadata (dict): The v05 plate metadata.

    Returns:
        NgioPlateMeta: The ngio plate metadata.
    """
    v05_plate = HCSV05(**metadata).ome.plate.model_dump()
    return NgioPlateMeta(plate=v05_plate, version="0.5")  # type: ignore


def ngio_to_v05_well_meta(metadata: NgioWellMeta) -> dict:
    """Convert a ngio well metadata to a v05 well metadata.

    Args:
        metadata (NgioWellMeta): The ngio well metadata.

    Returns:
        dict: The v05 well metadata.
    """
    v05_well = WellAttrsV05(well=metadata.model_dump())  # type: ignore
    v05_well = WellV05(ome=v05_well)
    return v05_well.model_dump(exclude_none=True, by_alias=True)


def ngio_to_v05_plate_meta(metadata: NgioPlateMeta) -> dict:
    """Convert a ngio plate metadata to a v05 plate metadata.

    Args:
        metadata (NgioPlateMeta): The ngio plate metadata.

    Returns:
        dict: The v05 plate metadata.
    """
    v05_plate = HCSAttrsV05(**metadata.model_dump())
    v05_plate = HCSV05(ome=v05_plate)
    return v05_plate.model_dump(exclude_none=True, by_alias=True)


class LabelsV05(BaseModel):
    ome: LabelsAttrsV05


def v05_to_ngio_labels_group_meta(
    metadata: dict,
) -> NgioLabelsGroupMeta:
    """Convert a v04 label group metadata to a ngio label group metadata.

    Args:
        metadata (dict): The v04 label group metadata.

    Returns:
        NgioLabelGroupMeta: The ngio label group metadata.
    """
    v05_label_group = LabelsV05(**metadata)
    return NgioLabelsGroupMeta(labels=v05_label_group.ome.labels, version="0.5")


def ngio_to_v05_labels_group_meta(metadata: NgioLabelsGroupMeta) -> dict:
    """Convert a ngio label group metadata to a v05 label group metadata.

    Args:
        metadata (NgioLabelsGroupMeta): The ngio label group metadata.

    Returns:
        dict: The v05 label group metadata.
    """
    v05_labels_attrs = LabelsAttrsV05(labels=metadata.labels, version="0.5")
    v05_labels_group = LabelsV05(ome=v05_labels_attrs)
    return v05_labels_group.model_dump(exclude_none=True, by_alias=True)



import json

from ome_zarr_models.v04.image import ImageAttrs as ImageAttrsV04
from ome_zarr_models.v04.image_label import ImageLabelAttrs as LabelAttrsV04
from ome_zarr_models.v04.well import WellAttrs as WellAttrsV04

from ngio.ome_zarr_meta import NgioImageMeta, NgioLabelMeta, NgioWellMeta
from ngio.ome_zarr_meta.v04._v04_spec import (
    ngio_to_v04_image_meta,
    ngio_to_v04_label_meta,
    ngio_to_v04_well_meta,
    v04_to_ngio_image_meta,
    v04_to_ngio_label_meta,
    v04_to_ngio_well_meta,
)


def test_image_round_trip():
    path = "tests/data/v04/meta/base_ome_zarr_image_meta.json"
    with open(path) as f:
        input_metadata = json.load(f)

    ngio_image = v04_to_ngio_image_meta(input_metadata)
    assert isinstance(ngio_image, NgioImageMeta)
    output_metadata = ngio_to_v04_image_meta(ngio_image)
    assert ImageAttrsV04(**output_metadata) == ImageAttrsV04(**input_metadata)


def test_label_round_trip():
    path = "tests/data/v04/meta/base_ome_zarr_label_meta.json"
    with open(path) as f:
        metadata = json.load(f)

    ngio_label = v04_to_ngio_label_meta(metadata)
    assert isinstance(ngio_label, NgioLabelMeta)
    output_metadata = ngio_to_v04_label_meta(ngio_label)
    assert LabelAttrsV04(**output_metadata) == LabelAttrsV04(**metadata)


def test_well_meta():
    path = "tests/data/v04/meta/base_ome_zarr_well_meta.json"
    with open(path) as f:
        metadata = json.load(f)

    ngio_well = v04_to_ngio_well_meta(metadata)
    assert isinstance(ngio_well, NgioWellMeta)
    output_metadata = ngio_to_v04_well_meta(ngio_well)
    assert isinstance(output_metadata, dict)
    assert WellAttrsV04(**output_metadata) == WellAttrsV04(**metadata)


def test_well_meta_path_normalization():
    path = "tests/data/v04/meta/ome_zarr_well_path_normalization_meta.json"
    with open(path) as f:
        metadata = json.load(f)

    ngio_well = v04_to_ngio_well_meta(metadata)
    assert isinstance(ngio_well, NgioWellMeta)
    output_metadata = ngio_to_v04_well_meta(ngio_well)
    assert isinstance(output_metadata, dict)

    images = [image["path"] for image in output_metadata["well"]["images"]]
    assert images == ["0", "0_mip"]



"""Implementation of the ROI Table class.

This class follows the roi_table specification at:
https://fractal-analytics-platform.github.io/fractal-tasks-core/tables/
"""

import warnings
from collections.abc import Iterable
from typing import Literal
from uuid import uuid4

import pandas as pd
from pydantic import BaseModel

from ngio.common import Roi
from ngio.tables._abstract_table import (
    AbstractBaseTable,
    TabularData,
)
from ngio.tables.backends import (
    BackendMeta,
    TableBackend,
    convert_to_pandas,
    normalize_pandas_df,
)
from ngio.utils import (
    NgioTableValidationError,
    NgioValueError,
    ZarrGroupHandler,
)

REQUIRED_COLUMNS = [
    "x_micrometer",
    "y_micrometer",
    "z_micrometer",
    "len_x_micrometer",
    "len_y_micrometer",
    "len_z_micrometer",
]

#####################
# Optional columns are not validated at the moment
# only a warning is raised if non optional columns are present
#####################

TIME_COLUMNS = [
    "t_second",
    "len_t_second",
]

ORIGIN_COLUMNS = [
    "x_micrometer_original",
    "y_micrometer_original",
    "z_micrometer_original",
]

TRANSLATION_COLUMNS = ["translation_x", "translation_y", "translation_z"]

PLATE_COLUMNS = [
    "plate_name",
    "row",
    "column",
    "path_in_well",
    "path_in_plate",
    "acquisition_id",
    "acquisition_name",
]

INDEX_COLUMNS = [
    "FieldIndex",
    "label",
]

OPTIONAL_COLUMNS = ORIGIN_COLUMNS + TRANSLATION_COLUMNS + PLATE_COLUMNS + INDEX_COLUMNS


def _check_optional_columns(col_name: str) -> None:
    """Check if the column name is in the optional columns."""
    if col_name not in OPTIONAL_COLUMNS + TIME_COLUMNS:
        warnings.warn(
            f"Column {col_name} is not in the optional columns.", stacklevel=2
        )


def _dataframe_to_rois(
    dataframe: pd.DataFrame,
    required_columns: list[str] = REQUIRED_COLUMNS,
) -> dict[str, Roi]:
    """Convert a DataFrame to a WorldCooROI object."""
    # Validate the columns of the DataFrame
    _missing_columns = set(required_columns).difference(set(dataframe.columns))
    if len(_missing_columns) != 0:
        raise NgioTableValidationError(
            f"Could not find required columns: {_missing_columns} in the table."
        )

    extra_columns = set(dataframe.columns).difference(
        set(required_columns + TIME_COLUMNS)
    )

    for col in extra_columns:
        _check_optional_columns(col)

    label_is_index = True if dataframe.index.name == "label" else False

    extras = {}

    rois = {}
    for row in dataframe.itertuples(index=True):
        # check if optional columns are present
        if len(extra_columns) > 0:
            extras = {col: getattr(row, col, None) for col in extra_columns}

        z_micrometer = getattr(row, "z_micrometer", None)
        z_length_micrometer = getattr(row, "len_z_micrometer", None)

        t_second = getattr(row, "t_second", None)
        t_length_second = getattr(row, "len_t_second", None)

        if label_is_index:
            label = int(row.Index)  # type: ignore (type can not be known here, but should be castable to int)
        else:
            label = getattr(row, "label", None)

        slices = {
            "x": (row.x_micrometer, row.len_x_micrometer),
            "y": (row.y_micrometer, row.len_y_micrometer),
            "z": (z_micrometer, z_length_micrometer),
        }
        if t_second is not None or t_length_second is not None:
            slices["t"] = (t_second, t_length_second)
        roi = Roi.from_values(
            name=str(row.Index),
            slices=slices,
            space="world",
            label=label,
            **extras,
        )
        rois[roi.name] = roi
    return rois


def _rois_to_dataframe(rois: dict[str, Roi], index_key: str | None) -> pd.DataFrame:
    """Convert a list of WorldCooROI objects to a DataFrame."""
    data = []
    for roi in rois.values():
        # This normalization is necessary for backward compatibility
        if roi.space != "world":
            raise NotImplementedError(
                "Only ROIs in world coordinates can be serialized."
            )

        z_slice = roi.get("z")
        if z_slice is None:
            z_micrometer = 0.0
            len_z_micrometer = 1.0
        else:
            z_micrometer = z_slice.start if z_slice.start is not None else 0.0
            len_z_micrometer = z_slice.length if z_slice.length is not None else 1.0

        x_slice = roi.get("x")
        if x_slice is None:
            raise NgioValueError("ROI is missing 'x' slice.")
        y_slice = roi.get("y")
        if y_slice is None:
            raise NgioValueError("ROI is missing 'y' slice.")
        row = {
            index_key: roi.get_name(),
            "x_micrometer": x_slice.start if x_slice.start is not None else 0.0,
            "y_micrometer": y_slice.start if y_slice.start is not None else 0.0,
            "z_micrometer": z_micrometer,
            "len_x_micrometer": x_slice.length if x_slice.length is not None else 1.0,
            "len_y_micrometer": y_slice.length if y_slice.length is not None else 1.0,
            "len_z_micrometer": len_z_micrometer,
        }

        t_slice = roi.get("t")
        if t_slice is not None:
            row["t_second"] = t_slice.start if t_slice.start is not None else 0.0
            row["len_t_second"] = t_slice.length if t_slice.length is not None else 1.0

        if roi.label is not None and index_key != "label":
            row["label"] = roi.label

        extra = roi.model_extra or {}
        for col in extra:
            _check_optional_columns(col)
            row[col] = extra[col]
        data.append(row)

    dataframe = pd.DataFrame(data)
    dataframe = normalize_pandas_df(dataframe, index_key=index_key)
    return dataframe


class RoiDictWrapper:
    """A wrapper for a dictionary of ROIs to provide a consistent interface."""

    def __init__(self, rois: Iterable[Roi]) -> None:
        self._rois_by_name = {}
        self._rois_by_label = {}
        for roi in rois:
            name = roi.name
            if name in self._rois_by_name:
                name = f"{name}_{uuid4().hex[:8]}"
            self._rois_by_name[name] = roi
            if roi.label is not None:
                self._rois_by_label[roi.label] = roi

    def get_by_name(self, name: str, default: Roi | None = None) -> Roi | None:
        """Get an ROI by its name."""
        return self._rois_by_name.get(name, default)

    def get_by_label(self, label: int, default: Roi | None = None) -> Roi | None:
        """Get an ROI by its label."""
        return self._rois_by_label.get(label, default)

    def _add_roi(self, roi: Roi, overwrite: bool = False) -> None:
        """Add an ROI to the wrapper."""
        if roi.name in self._rois_by_name and not overwrite:
            raise NgioValueError(f"ROI with name {roi.name} already exists.")

        self._rois_by_name[roi.name] = roi
        if roi.label is not None:
            self._rois_by_label[roi.label] = roi

    def add_rois(self, rois: Roi | Iterable[Roi], overwrite: bool = False) -> None:
        """Add ROIs to the wrapper."""
        if isinstance(rois, Roi):
            rois = [rois]

        for roi in rois:
            self._add_roi(roi, overwrite=overwrite)

    def to_list(self) -> list[Roi]:
        """Return the list of ROIs."""
        return list(self._rois_by_name.values())

    def to_dataframe(self, index_key: str | None = None) -> pd.DataFrame:
        """Convert the ROIs to a DataFrame."""
        return _rois_to_dataframe(self._rois_by_name, index_key=index_key)

    @classmethod
    def from_dataframe(
        cls, dataframe: pd.DataFrame, required_columns: list[str] = REQUIRED_COLUMNS
    ) -> "RoiDictWrapper":
        """Create a RoiDictWrapper from a DataFrame."""
        rois = _dataframe_to_rois(dataframe, required_columns=required_columns)
        return cls(rois.values())


def _table_to_rois(
    table: TabularData,
    index_key: str | None = None,
    index_type: Literal["int", "str"] | None = None,
    required_columns: list[str] = REQUIRED_COLUMNS,
) -> tuple[pd.DataFrame, RoiDictWrapper]:
    """Convert a table to a dictionary of ROIs.

    Args:
        table: The table to convert.
        index_key: The column name to use as the index of the DataFrame.
        index_type: The type of the index column in the DataFrame.
        required_columns: The required columns in the DataFrame.

    Returns:
        A tuple containing the DataFrame and a RoiDictWrapper with the ROIs.
    """
    dataframe = convert_to_pandas(
        table,
        index_key=index_key,
        index_type=index_type,
    )
    roi_dict_wrapper = RoiDictWrapper.from_dataframe(
        dataframe, required_columns=required_columns
    )
    return dataframe, roi_dict_wrapper


class GenericRoiTableV1(AbstractBaseTable):
    def __init__(
        self,
        *,
        rois: Iterable[Roi] | None = None,
        meta: BackendMeta,
        required_columns: list[str] = REQUIRED_COLUMNS,
    ) -> None:
        table = None

        self._rois: RoiDictWrapper | None = None
        if rois is not None:
            self._rois = RoiDictWrapper(rois)
            table = self._rois.to_dataframe(index_key=meta.index_key)

        self._required_columns = required_columns
        super().__init__(table_data=table, meta=meta)

    def __repr__(self) -> str:
        """Return a string representation of the table."""
        rois = self.rois()
        prop = f"num_rois={len(rois)}"
        class_name = self.__class__.__name__
        return f"{class_name}({prop})"

    @staticmethod
    def table_type() -> str:
        """Return the type of the table."""
        return "generic_roi_table"

    @staticmethod
    def version() -> Literal["1"]:
        """Return the version of the fractal table."""
        return "1"

    @property
    def table_data(self) -> TabularData:
        """Return the table."""
        if self._rois is None:
            return super().table_data

        if len(self.rois()) > 0:
            self._table_data = self._rois.to_dataframe(index_key=self.meta.index_key)
        return super().table_data

    def set_table_data(
        self, table_data: TabularData | None = None, refresh: bool = False
    ) -> None:
        if table_data is not None:
            if not isinstance(table_data, TabularData):
                raise NgioValueError(
                    "The table must be a pandas DataFrame, polars LazyFrame, "
                    " or AnnData object."
                )

            table_data, rois = _table_to_rois(
                table_data,
                index_key=self.index_key,
                index_type=self.index_type,
                required_columns=REQUIRED_COLUMNS,
            )
            self._table_data = table_data
            self._rois = rois
            return None

        if self._table_data is not None and not refresh:
            return None

        if self._table_backend is None:
            raise NgioValueError(
                "The table does not have a DataFrame in memory nor a backend."
            )

        table_data, rois = _table_to_rois(
            self._table_backend.load(),
            index_key=self.index_key,
            index_type=self.index_type,
            required_columns=REQUIRED_COLUMNS,
        )
        self._table_data = table_data
        self._rois = rois

    def _check_rois(self) -> None:
        """Load the ROIs from the table.

        If the ROIs are already loaded, do nothing.
        If the ROIs are not loaded, load them from the table.
        """
        if self._rois is None:
            self._rois = RoiDictWrapper.from_dataframe(
                self.dataframe, required_columns=self._required_columns
            )

    def rois(self) -> list[Roi]:
        """List all ROIs in the table."""
        self._check_rois()
        if self._rois is None:
            return []
        return self._rois.to_list()

    def add(self, roi: Roi | Iterable[Roi], overwrite: bool = False) -> None:
        """Append ROIs to the current table.

        Args:
            roi: A single ROI or a list of ROIs to add to the table.
            overwrite: If True, overwrite existing ROIs with the same name.
        """
        if isinstance(roi, Roi):
            roi = [roi]

        self._check_rois()
        if self._rois is None:
            self._rois = RoiDictWrapper([])

        self._rois.add_rois(roi, overwrite=overwrite)

    def get(self, roi_name: str) -> Roi:
        """Get an ROI from the table."""
        self._check_rois()
        if self._rois is None:
            self._rois = RoiDictWrapper([])

        roi = self._rois.get_by_name(roi_name)
        if roi is None:
            raise NgioValueError(f"ROI with name {roi_name} not found in the table.")
        return roi

    @classmethod
    def from_table_data(
        cls, table_data: TabularData, meta: BackendMeta
    ) -> "GenericRoiTableV1":
        """Create a new ROI table from a table data."""
        _, rois = _table_to_rois(
            table=table_data,
            index_key=meta.index_key,
            index_type=meta.index_type,
            required_columns=REQUIRED_COLUMNS,
        )
        return cls(rois=rois.to_list(), meta=meta)


class RoiTableV1Meta(BackendMeta):
    """Metadata for the ROI table."""

    table_version: Literal["1"] = "1"
    type: Literal["roi_table"] = "roi_table"
    index_key: str | None = "FieldIndex"
    index_type: Literal["str", "int"] | None = "str"


class RoiTableV1(GenericRoiTableV1):
    """Class to handle fractal ROI tables.

    To know more about the ROI table format, please refer to the
    specification at:
    https://fractal-analytics-platform.github.io/fractal-tasks-core/tables/
    """

    def __init__(
        self, rois: Iterable[Roi] | None = None, *, meta: RoiTableV1Meta | None = None
    ) -> None:
        """Create a new ROI table."""
        if meta is None:
            meta = RoiTableV1Meta()

        if meta.index_key is None:
            meta.index_key = "FieldIndex"

        if meta.index_type is None:
            meta.index_type = "str"
        super().__init__(meta=meta, rois=rois)

    @classmethod
    def from_handler(
        cls,
        handler: ZarrGroupHandler,
        backend: TableBackend | None = None,
    ) -> "RoiTableV1":
        table = cls._from_handler(
            handler=handler,
            backend=backend,
            meta_model=RoiTableV1Meta,
        )
        return table

    @staticmethod
    def table_type() -> Literal["roi_table"]:
        """Return the type of the table."""
        return "roi_table"


class RegionMeta(BaseModel):
    """Metadata for the region."""

    path: str


class MaskingRoiTableV1Meta(BackendMeta):
    """Metadata for the ROI table."""

    table_version: Literal["1"] = "1"
    type: Literal["masking_roi_table"] = "masking_roi_table"
    region: RegionMeta | None = None
    instance_key: str = "label"
    index_key: str | None = "label"
    index_type: Literal["int", "str"] | None = "int"


class MaskingRoiTableV1(GenericRoiTableV1):
    """Class to handle fractal ROI tables.

    To know more about the ROI table format, please refer to the
    specification at:
    https://fractal-analytics-platform.github.io/fractal-tasks-core/tables/
    """

    def __init__(
        self,
        rois: Iterable[Roi] | None = None,
        *,
        reference_label: str | None = None,
        meta: MaskingRoiTableV1Meta | None = None,
    ) -> None:
        """Create a new ROI table."""
        if meta is None:
            meta = MaskingRoiTableV1Meta()

        if reference_label is not None:
            path = f"../labels/{reference_label}"
            meta.region = RegionMeta(path=path)

        if meta.index_key is None:
            meta.index_key = "label"

        if meta.index_type is None:
            meta.index_type = "int"
        meta.instance_key = meta.index_key
        super().__init__(meta=meta, rois=rois)

    def __repr__(self) -> str:
        """Return a string representation of the table."""
        rois = self.rois()
        if self.reference_label is not None:
            prop = f"num_rois={len(rois)}, reference_label={self.reference_label}"
        else:
            prop = f"num_rois={len(rois)}"
        return f"MaskingRoiTableV1({prop})"

    @classmethod
    def from_handler(
        cls,
        handler: ZarrGroupHandler,
        backend: TableBackend | None = None,
    ) -> "MaskingRoiTableV1":
        table = cls._from_handler(
            handler=handler,
            backend=backend,
            meta_model=MaskingRoiTableV1Meta,
        )
        return table

    @staticmethod
    def table_type() -> Literal["masking_roi_table"]:
        """Return the type of the table."""
        return "masking_roi_table"

    @property
    def meta(self) -> MaskingRoiTableV1Meta:
        """Return the metadata of the table."""
        if not isinstance(self._meta, MaskingRoiTableV1Meta):
            raise NgioValueError(
                "The metadata of the table is not of type MaskingRoiTableV1Meta."
            )
        return self._meta

    @property
    def reference_label(self) -> str | None:
        """Return the reference label."""
        path = self.meta.region
        if path is None:
            return None

        path = path.path
        path = path.split("/")[-1]
        return path

    def get_label(self, label: int) -> Roi:
        """Get an ROI by label."""
        self._check_rois()
        if self._rois is None:
            self._rois = RoiDictWrapper([])
        roi = self._rois.get_by_label(label)

        if roi is None:
            raise NgioValueError(f"ROI with label {label} not found.")
        return roi



"""Fractal internal module for dataset metadata handling."""

from collections.abc import Sequence

from ngio.ome_zarr_meta.ngio_specs._axes import (
    AxesHandler,
)
from ngio.ome_zarr_meta.ngio_specs._pixel_size import PixelSize
from ngio.utils import NgioValidationError


class Dataset:
    """Model for a dataset in the multiscale."""

    def __init__(
        self,
        *,
        # args coming from ngff specs
        path: str,
        axes_handler: AxesHandler,
        scale: Sequence[float],
        translation: Sequence[float] | None = None,
    ):
        """Initialize the Dataset object.

        Args:
            path (str): The path of the dataset.
            axes_handler (AxesHandler): The axes handler object.
            scale (list[float]): The list of scale transformation.
                The scale transformation must have the same length as the axes.
            translation (list[float] | None): The list of translation.
                The translation must have the same length as the axes.
        """
        self._path = path
        self._axes_handler = axes_handler

        if len(scale) != len(axes_handler.axes):
            raise NgioValidationError(
                "The length of the scale transformation must be the same as the axes."
            )
        self._scale = list(scale)

        translation = translation or [0.0] * len(axes_handler.axes)
        if len(translation) != len(axes_handler.axes):
            raise NgioValidationError(
                "The length of the translation must be the same as the axes."
            )
        self._translation = list(translation)

    @property
    def path(self) -> str:
        """Return the path of the dataset."""
        return self._path

    @property
    def axes_handler(self) -> AxesHandler:
        """Return the axes handler object."""
        return self._axes_handler

    @property
    def pixel_size(self) -> PixelSize:
        """Return the pixel size for the dataset."""
        scale = self._scale
        pix_size_dict = {}
        # Mandatory axes: x, y
        for ax in ["x", "y"]:
            index = self.axes_handler.get_index(ax)
            assert index is not None
            pix_size_dict[ax] = scale[index]

        for ax in ["z", "t"]:
            index = self.axes_handler.get_index(ax)
            pix_size_dict[ax] = scale[index] if index is not None else 1.0

        return PixelSize(
            **pix_size_dict,
            space_unit=self.axes_handler.space_unit,
            time_unit=self.axes_handler.time_unit,
        )

    @property
    def scale(self) -> tuple[float, ...]:
        """Return the scale transformation as a tuple."""
        return tuple(self._scale)

    @property
    def translation(self) -> tuple[float, ...]:
        """Return the translation as a tuple."""
        return tuple(self._translation)



"""Common utilities for working with Zarr groups in consistent ways."""

import json
import warnings
from pathlib import Path
from typing import Literal

import dask.array as da
import fsspec
import zarr
from filelock import BaseFileLock, FileLock
from pydantic_zarr.v2 import ArraySpec as AnyArraySpecV2
from pydantic_zarr.v3 import ArraySpec as AnyArraySpecV3
from zarr.abc.store import Store
from zarr.errors import ContainsGroupError
from zarr.storage import FsspecStore, LocalStore, MemoryStore, ZipStore

from ngio.utils._cache import NgioCache
from ngio.utils._errors import (
    NgioFileExistsError,
    NgioFileNotFoundError,
    NgioValueError,
)

AccessModeLiteral = Literal["r", "r+", "w", "w-", "a"]
# StoreLike is more restrictive than it could be
# but to make sure we can handle the store correctly
# we need to be more restrictive
NgioSupportedStore = (
    str
    | Path
    | fsspec.mapping.FSMap
    | FsspecStore
    | MemoryStore
    | LocalStore
    | ZipStore
)
GenericStore = Store | NgioSupportedStore
StoreOrGroup = GenericStore | zarr.Group


def _check_store(store) -> NgioSupportedStore:
    """Check the store and return a valid store."""
    if not isinstance(store, NgioSupportedStore):
        warnings.warn(
            f"Store type {type(store)} is not explicitly supported. "
            f"Supported types are: {NgioSupportedStore}. "
            "Proceeding, but this may lead to unexpected behavior.",
            UserWarning,
            stacklevel=2,
        )
    return store


def _check_group(
    group: zarr.Group, mode: AccessModeLiteral | None = None
) -> zarr.Group:
    """Check the group and return a valid group."""
    if group.read_only and mode not in [None, "r"]:
        raise NgioValueError(f"The group is read only. Cannot open in mode {mode}.")

    if mode == "r" and not group.read_only:
        # let's make sure we don't accidentally write to the group
        group = zarr.open_group(store=group.store, path=group.path, mode="r")
    return group


def open_group_wrapper(
    store: StoreOrGroup,
    mode: AccessModeLiteral | None = None,
    zarr_format: Literal[2, 3] | None = None,
) -> zarr.Group:
    """Wrapper around zarr.open_group with some additional checks.

    Args:
        store (StoreOrGroup): The store or group to open.
        mode (AccessModeLiteral): The mode to open the group in.
        zarr_format (int): The Zarr format version to use.

    Returns:
        zarr.Group: The opened Zarr group.
    """
    if isinstance(store, zarr.Group):
        group = _check_group(store, mode)
        _check_store(group.store)
        return group

    try:
        _check_store(store)
        mode = mode if mode is not None else "a"
        group = zarr.open_group(store=store, mode=mode, zarr_format=zarr_format)

    except FileExistsError as e:
        raise NgioFileExistsError(
            f"A Zarr group already exists at {store}, consider setting overwrite=True."
        ) from e

    except FileNotFoundError as e:
        raise NgioFileNotFoundError(f"No Zarr group found at {store}") from e

    except ContainsGroupError as e:
        raise NgioFileExistsError(
            f"A Zarr group already exists at {store}, consider setting overwrite=True."
        ) from e

    return group


class ZarrGroupHandler:
    """A simple wrapper around a Zarr group to handle metadata."""

    def __init__(
        self,
        store: StoreOrGroup,
        zarr_format: Literal[2, 3] | None = None,
        cache: bool = False,
        mode: AccessModeLiteral | None = None,
    ):
        """Initialize the handler.

        Args:
            store (StoreOrGroup): The Zarr store or group containing the image data.
            meta_mode (str): The mode of the metadata handler.
            zarr_format (int | None): The Zarr format version to use.
            cache (bool): Whether to cache the metadata.
            mode (str | None): The mode of the store.
        """
        if mode not in ["r", "r+", "w", "w-", "a", None]:
            raise NgioValueError(f"Mode {mode} is not supported.")

        group = open_group_wrapper(store=store, mode=mode, zarr_format=zarr_format)
        self._group = group
        self.use_cache = cache

        self._group_cache: NgioCache[zarr.Group] = NgioCache(use_cache=cache)
        self._array_cache: NgioCache[zarr.Array] = NgioCache(use_cache=cache)
        self._handlers_cache: NgioCache[ZarrGroupHandler] = NgioCache(use_cache=cache)
        self._lock: tuple[Path, BaseFileLock] | None = None

    def __repr__(self) -> str:
        """Return a string representation of the handler."""
        return (
            f"ZarrGroupHandler(full_url={self.full_url}, read_only={self.read_only}, "
            f"cache={self.use_cache}"
        )

    @property
    def store(self) -> Store:
        """Return the store of the group."""
        return self._group.store

    @property
    def full_url(self) -> str | None:
        """Return the store path."""
        if isinstance(self.store, LocalStore):
            return (self.store.root / self.group.path).as_posix()
        elif isinstance(self.store, FsspecStore):
            return f"{self.store.path}/{self.group.path}"
        elif isinstance(self.store, ZipStore):
            return (self.store.path / self.group.path).as_posix()
        elif isinstance(self.store, MemoryStore):
            return None
        warnings.warn(
            f"Cannot determine full URL for store type {type(self.store)}. ",
            UserWarning,
            stacklevel=2,
        )
        return None

    @property
    def zarr_format(self) -> Literal[2, 3]:
        """Return the Zarr format version."""
        return self._group.metadata.zarr_format

    @property
    def read_only(self) -> bool:
        """Return whether the group is read only."""
        return self._group.read_only

    def _create_lock(self) -> tuple[Path, BaseFileLock]:
        """Create the lock."""
        if self._lock is not None:
            return self._lock

        if self.use_cache is True:
            raise NgioValueError(
                "Lock mechanism is not compatible with caching. "
                "Please set cache=False to use the lock mechanism."
            )

        if not isinstance(self.store, LocalStore):
            raise NgioValueError(
                "The store needs to be a LocalStore to use the lock mechanism. "
                f"Instead, got {self.store.__class__.__name__}."
            )

        store_path = Path(self.store.root) / self.group.path
        _lock_path = store_path.with_suffix(".lock")
        _lock = FileLock(_lock_path, timeout=10)
        return _lock_path, _lock

    @property
    def lock(self) -> BaseFileLock:
        """Return the lock."""
        if self._lock is None:
            self._lock = self._create_lock()
        return self._lock[1]

    @property
    def lock_path(self) -> Path:
        """Return the lock path."""
        if self._lock is None:
            self._lock = self._create_lock()
        return self._lock[0]

    def remove_lock(self) -> None:
        """Return the lock."""
        if self._lock is None:
            return None

        lock_path, lock = self._lock
        if lock_path.exists() and lock.lock_counter == 0:
            lock_path.unlink()
            self._lock = None
            return None

        raise NgioValueError("The lock is still in use. Cannot remove it.")

    def reopen_group(self) -> zarr.Group:
        """Reopen the group.

        This is useful when the group has been modified
        outside of the handler.
        """
        mode = "r" if self.read_only else "r+"
        return zarr.open_group(
            store=self._group.store,
            path=self._group.path,
            mode=mode,
            zarr_format=self._group.metadata.zarr_format,
        )

    def reopen_handler(self) -> "ZarrGroupHandler":
        """Reopen the handler.

        This is useful when the group has been modified
        outside of the handler.
        """
        mode = "r" if self.read_only else "r+"
        group = self.reopen_group()
        return ZarrGroupHandler(
            store=group,
            zarr_format=group.metadata.zarr_format,
            cache=self.use_cache,
            mode=mode,
        )

    def clean_cache(self) -> None:
        """Clear the cached metadata."""
        group = self.reopen_group()
        self.__init__(
            store=group,
            zarr_format=group.metadata.zarr_format,
            cache=self.use_cache,
            mode="r" if self.read_only else "r+",
        )

    @property
    def group(self) -> zarr.Group:
        """Return the group."""
        if self.use_cache is False:
            # If we are not using cache, we need to reopen the group
            # to make sure that the attributes are up to date
            return self.reopen_group()
        return self._group

    def load_attrs(self) -> dict:
        """Load the attributes of the group."""
        return self.reopen_group().attrs.asdict()

    def write_attrs(self, attrs: dict, overwrite: bool = False) -> None:
        """Write the metadata to the store."""
        # Maybe we should use the lock here
        if self.read_only:
            raise NgioValueError("The group is read only. Cannot write metadata.")
        group = self.reopen_group()
        if overwrite:
            group.attrs.clear()
        group.attrs.update(attrs)

    def create_group(self, path: str, overwrite: bool = False) -> zarr.Group:
        """Create a group in the group."""
        if self.group.read_only:
            raise NgioValueError("Cannot create a group in read only mode.")

        try:
            group = self.group.create_group(path, overwrite=overwrite)
        except ContainsGroupError as e:
            raise NgioFileExistsError(
                f"A Zarr group already exists at {path}, "
                "consider setting overwrite=True."
            ) from e
        self._group_cache.set(path, group, overwrite=overwrite)
        return group

    def get_group(
        self,
        path: str,
        create_mode: bool = False,
        overwrite: bool = False,
    ) -> zarr.Group:
        """Get a group from the group.

        Args:
            path (str): The path to the group.
            create_mode (bool): If True, create the group if it does not exist.
            overwrite (bool): If True, overwrite the group if it exists.

        Returns:
            zarr.Group: The Zarr group.

        """
        if overwrite and not create_mode:
            raise NgioValueError("Cannot overwrite a group without create_mode=True.")

        if overwrite:
            return self.create_group(path, overwrite=overwrite)

        group = self._group_cache.get(path)
        if isinstance(group, zarr.Group):
            return group

        group = self.group.get(path, default=None)
        if isinstance(group, zarr.Group):
            self._group_cache.set(path, group, overwrite=overwrite)
            return group

        if isinstance(group, zarr.Array):
            raise NgioValueError(f"The object at {path} is not a group, but an array.")

        if not create_mode:
            raise NgioFileNotFoundError(f"No group found at {path}")
        group = self.create_group(path)
        self._group_cache.set(path, group, overwrite=overwrite)
        return group

    def get_array(self, path: str) -> zarr.Array:
        """Get an array from the group."""
        array = self._array_cache.get(path)
        if isinstance(array, zarr.Array):
            return array
        array = self.group.get(path, default=None)
        if isinstance(array, zarr.Array):
            self._array_cache.set(path, array)
            return array

        if isinstance(array, zarr.Group):
            raise NgioValueError(f"The object at {path} is not an array, but a group.")
        raise NgioFileNotFoundError(f"No array found at {path}")

    def get_handler(
        self,
        path: str,
        create_mode: bool = True,
        overwrite: bool = False,
    ) -> "ZarrGroupHandler":
        """Get a new handler for a group in the current handler group.

        Args:
            path (str): The path to the group.
            create_mode (bool): If True, create the group if it does not exist.
            overwrite (bool): If True, overwrite the group if it exists.
        """
        handler = self._handlers_cache.get(path)
        if handler is not None:
            return handler
        group = self.get_group(path, create_mode=create_mode, overwrite=overwrite)
        mode = "r" if group.read_only else "r+"
        handler = ZarrGroupHandler(
            store=group, zarr_format=self.zarr_format, cache=self.use_cache, mode=mode
        )
        self._handlers_cache.set(path, handler)
        return handler

    @property
    def is_listable(self) -> bool:
        return is_group_listable(self.group)

    def delete_group(self, path: str) -> None:
        """Delete a group from the current group.

        Args:
            path (str): The path to the group to delete.
        """
        if self.group.read_only:
            raise NgioValueError("Cannot delete a group in read only mode.")
        self.group.__delitem__(path)
        self._group_cache._cache.pop(path, None)
        self._handlers_cache._cache.pop(path, None)

    def copy_group(self, dest_group: zarr.Group):
        """Copy the group to a new store."""
        copy_group(self.group, dest_group)


def find_dimension_separator(array: zarr.Array) -> Literal[".", "/"]:
    """Find the dimension separator used in the Zarr store.

    Args:
        array (zarr.Array): The Zarr array to check.

    Returns:
        Literal[".", "/"]: The dimension separator used in the store.
    """
    from zarr.core.chunk_key_encodings import DefaultChunkKeyEncoding

    if array.metadata.zarr_format == 2:
        separator = array.metadata.dimension_separator
    else:
        separator = array.metadata.chunk_key_encoding
        if not isinstance(separator, DefaultChunkKeyEncoding):
            raise NgioValueError(
                "Only DefaultChunkKeyEncoding is supported in this example."
            )
        separator = separator.separator
    return separator


def is_group_listable(group: zarr.Group) -> bool:
    """Check if a Zarr group is listable.

    A group is considered listable if it contains at least one array or subgroup.

    Args:
        group (zarr.Group): The Zarr group to check.

    Returns:
        bool: True if the group is listable, False otherwise.
    """
    if not group.store.supports_listing:
        # If the store does not support listing
        # then for sure it is not listable
        return False
    try:
        next(group.keys())
        return True
    except StopIteration:
        # Group is listable but empty
        return True
    except Exception as _:
        # Some stores may raise errors when listing
        # consider those not listable
        return False


def _make_sync_fs(fs: fsspec.AbstractFileSystem) -> fsspec.AbstractFileSystem:
    fs_dict = json.loads(fs.to_json())
    fs_dict["asynchronous"] = False
    return fsspec.AbstractFileSystem.from_json(json.dumps(fs_dict))


def _get_mapper(store: LocalStore | FsspecStore, path: str):
    if isinstance(store, LocalStore):
        fs = fsspec.filesystem("file")
        full_path = (store.root / path).as_posix()
    else:
        fs = _make_sync_fs(store.fs)
        full_path = f"{store.path}/{path}"
    return fs.get_mapper(full_path)


def _fsspec_copy(
    src_fs: LocalStore | FsspecStore,
    src_path: str,
    dest_fs: LocalStore | FsspecStore,
    dest_path: str,
):
    src_mapper = _get_mapper(src_fs, src_path)
    dest_mapper = _get_mapper(dest_fs, dest_path)
    for key in src_mapper.keys():
        dest_mapper[key] = src_mapper[key]


def _zarr_python_copy(src_group: zarr.Group, dest_group: zarr.Group):
    # Copy attributes
    dest_group.attrs.put(src_group.attrs.asdict())
    # Copy arrays
    for name, array in src_group.arrays():
        if array.metadata.zarr_format == 2:
            spec = AnyArraySpecV2.from_zarr(array)
        else:
            spec = AnyArraySpecV3.from_zarr(array)
        dst = spec.to_zarr(
            store=dest_group.store,
            path=f"{dest_group.path}/{name}",
            overwrite=True,
        )
        if array.ndim > 0:
            dask_array = da.from_zarr(array)
            da.to_zarr(dask_array, dst, overwrite=False)
    # Copy subgroups
    for name, subgroup in src_group.groups():
        dest_subgroup = dest_group.create_group(name, overwrite=True)
        _zarr_python_copy(subgroup, dest_subgroup)


def copy_group(
    src_group: zarr.Group, dest_group: zarr.Group, suppress_warnings: bool = False
):
    if src_group.metadata.zarr_format != dest_group.metadata.zarr_format:
        raise NgioValueError(
            "Different Zarr format versions between source and destination, "
            "cannot copy."
        )

    if not is_group_listable(src_group):
        raise NgioValueError("Source group is not listable, cannot copy.")

    if dest_group.read_only:
        raise NgioValueError("Destination group is read only, cannot copy.")
    if isinstance(src_group.store, LocalStore | FsspecStore) and isinstance(
        dest_group.store, LocalStore | FsspecStore
    ):
        _fsspec_copy(src_group.store, src_group.path, dest_group.store, dest_group.path)
        return
    if not suppress_warnings:
        warnings.warn(
            "Fsspec copy not possible, falling back to Zarr Python API for the copy. "
            "This will preserve some tabular data non-zarr native (parquet, and csv), "
            "and it will be slower for large datasets.",
            UserWarning,
            stacklevel=2,
        )
    _zarr_python_copy(src_group, dest_group)



from pathlib import Path

import numpy as np
import pytest
from pydantic import ValidationError
from zarr.storage import MemoryStore

from ngio import (
    OmeZarrContainer,
    create_empty_ome_zarr,
    create_ome_zarr_from_array,
    create_synthetic_ome_zarr,
)
from ngio.images._create_utils import init_image_like_from_shapes
from ngio.ome_zarr_meta import NgioImageMeta
from ngio.utils import NgioValueError


@pytest.mark.parametrize(
    "create_kwargs",
    [
        {
            "store": "test_image_yx._zarr",
            "shape": (64, 64),
            "xy_pixelsize": 0.5,
            "axes_names": ["y", "x"],
        },
        {
            "store": "test_image_cyx.zarr",
            "shape": (2, 64, 64),
            "xy_pixelsize": 0.5,
            "axes_names": ["c", "y", "x"],
            "channel_labels": ["channel1", "channel2"],
        },
        {
            "store": "test_image_zyx.zarr",
            "shape": (3, 64, 64),
            "xy_pixelsize": 0.5,
            "z_spacing": 2.0,
            "axes_names": ["z", "y", "x"],
        },
        {
            "store": "test_image_czyx.zarr",
            "shape": (2, 3, 64, 64),
            "xy_pixelsize": 0.5,
            "z_spacing": 2.0,
            "axes_names": ["c", "z", "y", "x"],
            "channel_labels": ["channel1", "channel2"],
        },
        {
            "store": "test_image_c1yx.zarr",
            "shape": (2, 1, 64, 64),
            "xy_pixelsize": 0.5,
            "z_spacing": 1.0,
            "axes_names": ["c", "z", "y", "x"],
            "channel_labels": ["channel1", "channel2"],
        },
        {
            "store": "test_image_tyx.zarr",
            "shape": (4, 64, 64),
            "xy_pixelsize": 0.5,
            "time_spacing": 4.0,
            "axes_names": ["t", "y", "x"],
        },
        {
            "store": "test_image_tcyx.zarr",
            "shape": (4, 2, 64, 64),
            "xy_pixelsize": 0.5,
            "time_spacing": 4.0,
            "axes_names": ["t", "c", "y", "x"],
            "channel_labels": ["channel1", "channel2"],
        },
        {
            "store": "test_image_tzyx.zarr",
            "shape": (4, 3, 64, 64),
            "xy_pixelsize": 0.5,
            "z_spacing": 2.0,
            "time_spacing": 4.0,
            "axes_names": ["t", "z", "y", "x"],
        },
        {
            "store": "test_image_tczyx.zarr",
            "shape": (4, 2, 3, 64, 64),
            "xy_pixelsize": 0.5,
            "z_spacing": 2.0,
            "time_spacing": 4.0,
            "axes_names": ["t", "c", "z", "y", "x"],
            "channel_labels": ["channel1", "channel2"],
        },
    ],
)
def test_create_empty(tmp_path: Path, create_kwargs: dict):
    create_kwargs["store"] = tmp_path / create_kwargs["store"]
    ome_zarr = create_empty_ome_zarr(**create_kwargs, dtype="uint8", levels=1)
    ome_zarr.derive_label("label1")

    shape = create_kwargs.pop("shape")
    array = np.random.randint(0, 255, shape, dtype="uint8")
    create_ome_zarr_from_array(array=array, **create_kwargs, levels=1, overwrite=True)

    store = create_kwargs["store"]
    create_synthetic_ome_zarr(store=store, shape=shape, overwrite=True)


def test_large_synt(tmp_path: Path):
    store = tmp_path / "test_image_large.zarr"
    create_synthetic_ome_zarr(store=store, shape=(1, 1, 513, 513), overwrite=True)


def test_create_fail(tmp_path: Path):
    with pytest.raises(NgioValueError):
        create_ome_zarr_from_array(
            array=np.random.randint(0, 255, (64, 64), dtype="uint8"),
            store=tmp_path / "fail.zarr",
            xy_pixelsize=0.5,
            axes_names=["z", "y", "x"],  # should fail expected yx
            levels=1,
            overwrite=True,
        )

    with pytest.raises(NgioValueError):
        create_ome_zarr_from_array(
            array=np.random.randint(0, 255, (2, 64, 64), dtype="uint8"),
            store=tmp_path / "fail.zarr",
            xy_pixelsize=0.5,
            axes_names=["c", "y", "x"],
            levels=1,
            channel_labels=[
                "channel1",
                "channel2",
                "channel3",
            ],  # should fail expected 2 channels
            overwrite=True,
        )

    with pytest.raises(ValidationError):
        create_ome_zarr_from_array(
            array=np.random.randint(0, 255, (2, 64, 64), dtype="uint8"),
            store=tmp_path / "fail.zarr",
            xy_pixelsize=0.5,
            axes_names=["c", "y", "x"],
            levels=1,
            chunks=(1, 64, 64, 64),  # should fail expected 3 axes
            overwrite=True,
        )


def test_derive_label_channels_policy():
    store = MemoryStore()
    ome_zarr = create_synthetic_ome_zarr(store, shape=(3, 1, 64, 65))

    label = ome_zarr.derive_label("test-label-singleton", channels_policy="singleton")
    assert label.dimensions.get("c") == 1
    label = ome_zarr.derive_label("test-label-same", channels_policy="same")
    assert label.dimensions.get("c") == 3
    label = ome_zarr.derive_label("test-label-squeeze", channels_policy="squeeze")
    assert "c" not in label.axes
    assert label.dimensions.get("c") is None
    label = ome_zarr.derive_label("test-label-int", channels_policy=2)
    assert label.dimensions.get("c") == 2


def test_derive_from_non_dishogeneus_shapes():
    # Yes those shapes are intentionally weird
    shapes = [
        (4, 3, 64, 65),
        (4, 3, 64, 50),
        (4, 3, 32, 25),
    ]
    store = MemoryStore()
    image_handler = init_image_like_from_shapes(
        store=store,
        meta_type=NgioImageMeta,
        shapes=shapes,
        pixelsize=0.5,
    )
    ome_zarr = OmeZarrContainer(group_handler=image_handler)
    ome_zarr.derive_label("test-label-same", channels_policy="same")
    for path in ome_zarr.levels_paths:
        img = ome_zarr.get_image(path=path)
        lbl = ome_zarr.get_label(name="test-label-same", path=path)
        assert img.shape == lbl.shape

    image = ome_zarr.get_image(path="1")
    ome_zarr.derive_label("test-label-level-1", ref_image=image, channels_policy="same")

    for path_img, path_lbl in zip(["1", "2"], ["0", "1"], strict=True):
        img = ome_zarr.get_image(path=path_img)
        lbl = ome_zarr.get_label(name="test-label-level-1", path=path_lbl)
        assert img.shape == lbl.shape

    lbl = ome_zarr.get_label(name="test-label-level-1", path="2")
    scaling_factor = tuple(s1 / s2 for s1, s2 in zip(shapes[0], shapes[1], strict=True))
    assert image.meta.scaling_factor() == scaling_factor
    assert lbl.shape == (4, 3, 32, 19)


def test_fail_derive_singleton():
    store = MemoryStore()
    ome_zarr = create_empty_ome_zarr(store=store, shape=(1, 1, 64, 4), pixelsize=0.5)
    expected_shapes = [
        (1, 1, 64, 4),
        (1, 1, 32, 2),
        (1, 1, 16, 1),
        (1, 1, 8, 1),
        (1, 1, 4, 1),
    ]
    expected_pixel_size_x = [0.5, 1.0, 2.0, 2.0, 2.0]
    for path, shape, px_x in zip(
        ome_zarr.levels_paths, expected_shapes, expected_pixel_size_x, strict=True
    ):
        img = ome_zarr.get_image(path=path)
        assert img.shape == shape
        assert img.pixel_size.x == px_x


def test_fail_create_from_non_decreasing_shapes():
    # Yes those shapes are intentionally weird
    shapes = [
        (4, 3, 64, 64),
        (4, 3, 128, 128),
        (4, 3, 256, 256),
    ]
    store = MemoryStore()

    with pytest.raises(NgioValueError):
        _ = init_image_like_from_shapes(
            store=store,
            meta_type=NgioImageMeta,
            shapes=shapes,
            pixelsize=0.5,
        )



"""Image metadata models.

This module contains the models for the image metadata.
These metadata models are not adhering to the OME standard.
But they can be built from the OME standard metadata, and the
can be converted to the OME standard.
"""

from collections.abc import Sequence
from typing import Any, Literal, TypeVar

import numpy as np
from pydantic import BaseModel

from ngio.ome_zarr_meta.ngio_specs._axes import (
    AxesHandler,
    DefaultSpaceUnit,
    DefaultTimeUnit,
    SpaceUnits,
    TimeUnits,
)
from ngio.ome_zarr_meta.ngio_specs._channels import ChannelsMeta
from ngio.ome_zarr_meta.ngio_specs._dataset import Dataset
from ngio.ome_zarr_meta.ngio_specs._pixel_size import PixelSize
from ngio.utils import NgioValidationError, NgioValueError

T = TypeVar("T")
NgffVersions = Literal["0.4", "0.5"]
DefaultNgffVersion: Literal["0.4"] = "0.4"


class ImageLabelSource(BaseModel):
    """Image label source model."""

    version: NgffVersions
    source: dict[str, str | None]

    @classmethod
    def default_init(cls, version: NgffVersions) -> "ImageLabelSource":
        """Initialize the ImageLabelSource object."""
        return cls(version=version, source={"image": "../../"})


class NgioLabelsGroupMeta(BaseModel):
    """Metadata model for the /labels group in OME-NGFF."""

    version: NgffVersions
    labels: list[str]


class AbstractNgioImageMeta:
    """Base class for ImageMeta and LabelMeta."""

    def __init__(
        self, version: NgffVersions, name: str | None, datasets: list[Dataset]
    ) -> None:
        """Initialize the ImageMeta object."""
        self._version = version
        self._name = name

        if len(datasets) == 0:
            raise NgioValidationError("At least one dataset must be provided.")

        self._datasets = datasets
        self._axes_handler = datasets[0].axes_handler

    def __repr__(self):
        class_name = type(self).__name__
        paths = [dataset.path for dataset in self.datasets]
        axes = self.axes_handler.axes_names
        return f"{class_name}(name={self.name}, datasets={paths}, axes={axes})"

    @classmethod
    def default_init(
        cls,
        levels: Sequence[str],
        axes_handler: AxesHandler,
        scales: Sequence[tuple[float, ...]],
        translations: Sequence[tuple[float, ...] | None],
        name: str | None = None,
        version: NgffVersions = DefaultNgffVersion,
    ):
        """Initialize the ImageMeta object."""
        datasets = []
        for level, scale, translation in zip(levels, scales, translations, strict=True):
            dataset = Dataset(
                path=level,
                axes_handler=axes_handler,
                scale=scale,
                translation=translation,
            )
            datasets.append(dataset)

        return cls(
            version=version,
            name=name,
            datasets=datasets,
        )

    def to_units(
        self,
        *,
        space_unit: SpaceUnits = DefaultSpaceUnit,
        time_unit: TimeUnits = DefaultTimeUnit,
    ):
        """Convert the pixel size to the given units.

        Args:
            space_unit(str): The space unit to convert to.
            time_unit(str): The time unit to convert to.
        """
        new_axes_handler = self.axes_handler.to_units(
            space_unit=space_unit,
            time_unit=time_unit,
        )
        new_datasets = []
        for dataset in self.datasets:
            new_dataset = Dataset(
                path=dataset.path,
                axes_handler=new_axes_handler,
                scale=dataset.scale,
                translation=dataset.translation,
            )
            new_datasets.append(new_dataset)

        return type(self)(
            version=self.version,
            name=self.name,
            datasets=new_datasets,
        )

    @property
    def version(self) -> NgffVersions:
        """Version of the OME-NFF metadata used to build the object."""
        return self._version  # type: ignore (version is a Literal type)

    @property
    def zarr_format(self) -> Literal[2, 3]:
        """Zarr version used to store the data."""
        match self.version:
            case "0.4":
                return 2
            case "0.5":
                return 3
            case _:
                raise NgioValueError(f"Unsupported NGFF version: {self.version}")

    @property
    def name(self) -> str | None:
        """Name of the image."""
        return self._name

    @property
    def datasets(self) -> list[Dataset]:
        """List of datasets in the multiscale."""
        return self._datasets

    @property
    def axes_handler(self):
        """Return the axes mapper."""
        return self._axes_handler

    @property
    def levels(self) -> int:
        """Number of levels in the multiscale."""
        return len(self.datasets)

    @property
    def paths(self) -> list[str]:
        """List of paths of the datasets."""
        return [dataset.path for dataset in self.datasets]

    @property
    def space_unit(self) -> str | None:
        """Get the space unit of the pixel size."""
        return self.axes_handler.space_unit

    @property
    def time_unit(self) -> str | None:
        """Get the time unit of the pixel size."""
        return self.axes_handler.time_unit

    def _get_dataset_by_path(self, path: str) -> Dataset:
        """Get a dataset by its path."""
        for dataset in self.datasets:
            if dataset.path == path:
                return dataset
        raise NgioValueError(f"Dataset with path {path} not found.")

    def _get_dataset_by_index(self, idx: int) -> Dataset:
        """Get a dataset by its index."""
        if idx < 0 or idx >= len(self.datasets):
            raise NgioValueError(f"Index {idx} out of range.")
        return self.datasets[idx]

    def _find_closest_dataset(
        self, pixel_size: PixelSize, mode: str = "any"
    ) -> Dataset | None:
        """Find the closest dataset to the given pixel size.

        Args:
            pixel_size(PixelSize): The pixel size to search for.
            mode(str): The mode to find the closest dataset.
                "any": Will find the closest dataset.
                "lr": Will find closest "lower" resolution dataset.
                "hr": Will find closest "higher" resolution
        """
        min_dist = np.inf
        closest_dataset = None

        if mode == "any":
            datasets = self.datasets
        elif mode == "lr":
            # Lower resolution means that the pixel size is larger.
            datasets = [d for d in self.datasets if d.pixel_size > pixel_size]
        elif mode == "hr":
            # Higher resolution means that the pixel size is smaller.
            datasets = [d for d in self.datasets if d.pixel_size < pixel_size]
        else:
            raise NgioValueError(f"Mode {mode} not recognized.")

        for d in datasets:
            dist = d.pixel_size.distance(pixel_size)
            if dist < min_dist:
                min_dist = dist
                closest_dataset = d

        return closest_dataset

    def _get_closest_dataset(
        self, pixel_size: PixelSize, strict: bool = False
    ) -> Dataset:
        """Get a dataset with the closest pixel size.

        Args:
            pixel_size(PixelSize): The pixel size to search for.
            strict(bool): If True, the pixel size must be exactly the same.
        """
        closest_dataset = self._find_closest_dataset(pixel_size, mode="any")

        if closest_dataset is None:
            raise NgioValueError("No dataset found.")

        if strict and closest_dataset.pixel_size != pixel_size:
            raise NgioValueError(
                "No dataset with a pixel size close enough. "
                "Best match is "
                f"{closest_dataset.path}:{closest_dataset.pixel_size}"
            )
        return closest_dataset

    def get_lowest_resolution_dataset(self) -> Dataset:
        """Get the dataset with the lowest resolution."""
        dataset = self.datasets[-1]
        while True:
            lower_res_dataset = self._find_closest_dataset(
                dataset.pixel_size, mode="lr"
            )
            if lower_res_dataset is None:
                break
            dataset = lower_res_dataset
        return dataset

    def get_highest_resolution_dataset(self) -> Dataset:
        """Get the dataset with the highest resolution."""
        dataset = self.datasets[0]
        while True:
            higher_res_dataset = self._find_closest_dataset(
                dataset.pixel_size, mode="hr"
            )
            if higher_res_dataset is None:
                break
            dataset = higher_res_dataset
        return dataset

    def get_dataset(
        self,
        *,
        path: str | None = None,
        idx: int | None = None,
        pixel_size: PixelSize | None = None,
        strict: bool = False,
    ) -> Dataset:
        """Get a dataset by its path, index or pixel size.

        If all arguments are None, the dataset with the highest resolution is returned.

        Args:
            path(str): The path of the dataset.
            idx(int): The index of the dataset.
            pixel_size(PixelSize): The pixel size to search for.
            strict(bool): If True, the pixel size must be exactly the same.
                If pixel_size is None, strict is ignored.
        """
        # Only one of the arguments must be provided
        if (
            sum(
                [
                    path is not None,
                    idx is not None,
                    pixel_size is not None,
                ]
            )
            > 1
        ):
            raise NgioValueError("get_dataset must receive only one argument or None.")

        if path is not None:
            return self._get_dataset_by_path(path)
        elif idx is not None:
            return self._get_dataset_by_index(idx)
        elif pixel_size is not None:
            return self._get_closest_dataset(pixel_size, strict=strict)
        else:
            return self.get_highest_resolution_dataset()

    def _get_closest_datasets(self, path: str | None = None) -> tuple[Dataset, Dataset]:
        """Get the closest datasets to a dataset."""
        dataset = self.get_dataset(path=path)
        lr_dataset = self._find_closest_dataset(dataset.pixel_size, mode="lr")
        if lr_dataset is None:
            raise NgioValueError(
                "No lower resolution dataset found. "
                "This is the lowest resolution dataset."
            )
        return dataset, lr_dataset

    def scaling_factor(self, path: str | None = None) -> tuple[float, ...]:
        """Get the scaling factors to downscale to the next lower resolution dataset."""
        if self.levels == 1:
            return (1.0,) * len(self.axes_handler.axes_names)
        dataset, lr_dataset = self._get_closest_datasets(path=path)
        scale = dataset.scale
        lr_scale = lr_dataset.scale
        scaling_factors = [s / s_lr for s_lr, s in zip(scale, lr_scale, strict=True)]
        return tuple(scaling_factors)


class NgioLabelMeta(AbstractNgioImageMeta):
    """Label metadata model."""

    def __init__(
        self,
        version: NgffVersions,
        name: str | None,
        datasets: list[Dataset],
        image_label: ImageLabelSource | None = None,
    ) -> None:
        """Initialize the ImageMeta object."""
        super().__init__(version, name, datasets)
        image_label = (
            ImageLabelSource.default_init(self.version)
            if image_label is None
            else image_label
        )
        assert image_label is not None
        if image_label.version != version:
            raise NgioValidationError(
                "Label image version must match the metadata version."
            )
        self._image_label = image_label

    @property
    def source_image(self) -> str | None:
        source = self._image_label.source
        if "image" not in source:
            return None

        image_path = source["image"]
        return image_path

    @property
    def image_label(self) -> ImageLabelSource:
        """Get the image label metadata."""
        return self._image_label


class NgioImageMeta(AbstractNgioImageMeta):
    """Image metadata model."""

    def __init__(
        self,
        version: NgffVersions,
        name: str | None,
        datasets: list[Dataset],
        channels: ChannelsMeta | None = None,
    ) -> None:
        """Initialize the ImageMeta object."""
        super().__init__(version=version, name=name, datasets=datasets)
        self._channels_meta = channels

    @property
    def channels_meta(self) -> ChannelsMeta | None:
        """Get the channels_meta metadata."""
        return self._channels_meta

    def set_channels_meta(self, channels_meta: ChannelsMeta) -> None:
        """Set channels_meta metadata."""
        self._channels_meta = channels_meta

    def init_channels(
        self,
        labels: list[str] | int,
        wavelength_ids: list[str] | None = None,
        colors: list[str] | None = None,
        active: list[bool] | None = None,
        start: list[int | float] | None = None,
        end: list[int | float] | None = None,
        data_type: Any = np.uint16,
    ) -> None:
        """Set the channels_meta metadata for the image.

        Args:
            labels (list[str]|int): The labels of the channels.
            wavelength_ids (list[str], optional): The wavelengths of the channels.
            colors (list[str], optional): The colors of the channels.
            adjust_window (bool, optional): Whether to adjust the window.
            start_percentile (int, optional): The start percentile.
            end_percentile (int, optional): The end percentile.
            active (list[bool], optional): Whether the channel is active.
            start (list[int | float], optional): The start value of the channel.
            end (list[int | float], optional): The end value of the channel.
            end (int): The end value of the channel.
            data_type (Any): The data type of the channel.
        """
        channels_meta = ChannelsMeta.default_init(
            labels=labels,
            wavelength_id=wavelength_ids,
            colors=colors,
            active=active,
            start=start,
            end=end,
            data_type=data_type,
        )
        self.set_channels_meta(channels_meta=channels_meta)


NgioImageLabelMeta = NgioImageMeta | NgioLabelMeta



"""Various utilities for the ngio package."""

from ngio.utils._datasets import (
    download_ome_zarr_dataset,
    list_ome_zarr_datasets,
    print_datasets_infos,
)
from ngio.utils._errors import (
    NgioError,
    NgioFileExistsError,
    NgioFileNotFoundError,
    NgioTableValidationError,
    NgioValidationError,
    NgioValueError,
)
from ngio.utils._fractal_fsspec_store import fractal_fsspec_store
from ngio.utils._zarr_utils import (
    AccessModeLiteral,
    NgioCache,
    StoreOrGroup,
    ZarrGroupHandler,
    copy_group,
    open_group_wrapper,
)

__all__ = [
    "AccessModeLiteral",
    "NgioCache",
    "NgioError",
    "NgioFileExistsError",
    "NgioFileNotFoundError",
    "NgioTableValidationError",
    "NgioValidationError",
    "NgioValueError",
    "StoreOrGroup",
    "ZarrGroupHandler",
    "copy_group",
    "download_ome_zarr_dataset",
    "fractal_fsspec_store",
    "list_ome_zarr_datasets",
    "open_group_wrapper",
    "print_datasets_infos",
]



from functools import partial
from typing import Literal

import dask.array as da
import numpy as np
from scipy.ndimage import zoom as scipy_zoom

from ngio.utils import NgioValueError

InterpolationOrder = Literal["nearest", "linear", "cubic"]


def order_to_int(order: InterpolationOrder | Literal[0, 1, 2]) -> Literal[0, 1, 2]:
    if order == "nearest" or order == 0:
        return 0
    elif order == "linear" or order == 1:
        return 1
    elif order == "cubic" or order == 2:
        return 2
    else:
        raise NgioValueError(f"Invalid order: {order}")


def _stacked_zoom(x, zoom_y, zoom_x, order=1, mode="grid-constant", grid_mode=True):
    *rest, yshape, xshape = x.shape
    x = x.reshape(-1, yshape, xshape)
    scale_xy = (zoom_y, zoom_x)
    _x_out = [
        scipy_zoom(x[i], scale_xy, order=order, mode=mode, grid_mode=grid_mode)
        for i in range(x.shape[0])
    ]
    x_out = np.stack(_x_out)  # type: ignore (scipy_zoom returns np.ndarray, but type is not inferred correctly)
    return x_out.reshape(*rest, *x_out.shape[1:])


def fast_zoom(x, zoom, order=1, mode="grid-constant", grid_mode=True, auto_stack=True):
    """Fast zoom function.

    Scipy zoom function that can handle singleton dimensions
    but the performance degrades with the number of dimensions.

    This function has two small optimizations:
     - it removes singleton dimensions before calling zoom
     - if it detects that the zoom is only on the last two dimensions
         it stacks the first dimensions to call zoom only on the last two.
    """
    mask = np.isclose(x.shape, 1)
    # Always keep the last two dimensions
    # To avoid issues with singleton x or y dimensions
    mask[-1] = False
    mask[-2] = False
    zoom = np.array(zoom)
    singletons = tuple(np.where(mask)[0])
    xs = np.squeeze(x, axis=singletons)
    new_zoom = zoom[~mask]

    *zoom_rest, zoom_y, zoom_x = new_zoom
    if auto_stack and np.allclose(zoom_rest, 1):
        xs = _stacked_zoom(
            xs, zoom_y, zoom_x, order=order, mode=mode, grid_mode=grid_mode
        )
    else:
        xs = scipy_zoom(xs, new_zoom, order=order, mode=mode, grid_mode=grid_mode)
    x = np.expand_dims(xs, axis=singletons)  # type: ignore (scipy_zoom returns np.ndarray, but type is not inferred correctly)
    return x


def _zoom_inputs_check(
    source_array: np.ndarray | da.Array,
    scale: tuple[int | float, ...] | None = None,
    target_shape: tuple[int, ...] | None = None,
) -> tuple[np.ndarray, tuple[int, ...]]:
    if scale is None and target_shape is None:
        raise NgioValueError("Either scale or target_shape must be provided")

    if scale is not None and target_shape is not None:
        raise NgioValueError("Only one of scale or target_shape must be provided")

    if scale is None:
        assert target_shape is not None, "Target shape must be provided"
        if len(target_shape) != source_array.ndim:
            raise NgioValueError(
                "Target shape must have the "
                "same number of dimensions as "
                "the source array"
            )
        _scale = np.array(target_shape) / np.array(source_array.shape)
        _target_shape = target_shape
    else:
        _scale = np.array(scale)
        _target_shape = tuple(map(int, np.round(np.array(source_array.shape) * scale)))

    if len(_scale) != source_array.ndim:
        raise NgioValueError(
            f"Cannot scale array of shape {source_array.shape} with factors {_scale}."
            " Target shape must have the same number of dimensions as the source array."
        )

    return _scale, _target_shape


def dask_zoom(
    source_array: da.Array,
    scale: tuple[float | int, ...] | None = None,
    target_shape: tuple[int, ...] | None = None,
    order: InterpolationOrder = "linear",
) -> da.Array:
    """Dask implementation of zooming an array.

    Only one of scale or target_shape must be provided.

    Args:
        source_array (da.Array): The source array to zoom.
        scale (tuple[int, ...] | None): The scale factor to zoom by.
        target_shape (tuple[int, ...], None): The target shape to zoom to.
        order (Literal["nearest", "linear", "cubic"]): The order of interpolation.
            Defaults to "linear".

    Returns:
        da.Array: The zoomed array.
    """
    # This function follow the implementation from:
    # https://github.com/ome/ome-zarr-py/blob/master/ome_zarr/dask_utils.py
    # The module was contributed by Andreas Eisenbarth @aeisenbarth
    # See https://github.com/toloudis/ome-zarr-py/pull/
    _scale, _target_shape = _zoom_inputs_check(
        source_array=source_array, scale=scale, target_shape=target_shape
    )

    # Rechunk to better match the scaling operation
    source_chunks = np.array(source_array.chunksize)  # type: ignore (da.Array.chunksize is a tuple of ints)
    better_source_chunks = np.maximum(1, np.round(source_chunks * _scale) / _scale)
    better_source_chunks = better_source_chunks.astype(int)
    source_array = source_array.rechunk(better_source_chunks)  # type: ignore (better_source_chunks is a valid input for rechunk)

    # Calculate the block output shape
    block_output_shape = tuple(np.ceil(better_source_chunks * _scale).astype(int))

    zoom_wrapper = partial(
        fast_zoom,
        zoom=_scale,
        order=order_to_int(order),
        mode="grid-constant",
        grid_mode=True,
    )

    out_array = da.map_blocks(
        zoom_wrapper, source_array, chunks=block_output_shape, dtype=source_array.dtype
    )

    # Slice and rechunk to target
    slices = tuple(slice(0, ts, 1) for ts in _target_shape)
    out_array = out_array[slices]
    return out_array


def numpy_zoom(
    source_array: np.ndarray,
    scale: tuple[int | float, ...] | None = None,
    target_shape: tuple[int, ...] | None = None,
    order: InterpolationOrder = "linear",
) -> np.ndarray:
    """Numpy implementation of zooming an array.

    Only one of scale or target_shape must be provided.

    Args:
        source_array (np.ndarray): The source array to zoom.
        scale (tuple[int, ...] | None): The scale factor to zoom by.
        target_shape (tuple[int, ...], None): The target shape to zoom to.
        order (Literal[0, 1, 2]): The order of interpolation. Defaults to 1.

    Returns:
        np.ndarray: The zoomed array
    """
    _scale, _ = _zoom_inputs_check(
        source_array=source_array, scale=scale, target_shape=target_shape
    )

    out_array = fast_zoom(
        source_array,
        zoom=_scale,
        order=order_to_int(order),
        mode="grid-constant",
        grid_mode=True,
    )
    assert isinstance(out_array, np.ndarray)
    return out_array
